import os
import numba
import psutil
from numba.cuda import CudaSupportError
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QWidget, QVBoxLayout, QHBoxLayout, QLabel, QGroupBox

from aydin.util.misc.units import human_readable_byte_size


class SystemSummaryWidget(QWidget):
    def __init__(self, parent):
        QWidget.__init__(self, parent)

        self.layout = QHBoxLayout()
        self.layout.setAlignment(Qt.AlignCenter)

        # CPU summary
        self.cpu_group_box = QGroupBox("CPU Summary")
        self.cpu_group_box_layout = QVBoxLayout()
        self.cpu_group_box_layout.setSpacing(0)
        self.cpu_group_box_layout.setAlignment(Qt.AlignTop)
        self.cpu_group_box.setLayout(self.cpu_group_box_layout)

        # CPU freq
        self.cpu_freq_stats_label = QLabel(
            f"Current CPU frequency:\t {round(psutil.cpu_freq().current, 2)} Mhz", self
        )
        self.cpu_group_box_layout.addWidget(self.cpu_freq_stats_label)

        # Number of cores
        self.nb_cores_label = QLabel(
            f"Number of CPU cores:\t {os.cpu_count() // 2}", self
        )
        self.cpu_group_box_layout.addWidget(self.nb_cores_label)
        if (os.cpu_count() // 2) < 4:
            self.nb_cores_label.setStyleSheet("QLabel {color: red;}")
        elif (os.cpu_count() // 2) <= 6:
            self.nb_cores_label.setStyleSheet("QLabel {color: orange;}")
        else:
            self.nb_cores_label.setStyleSheet("QLabel {color: green;}")

        self.cpu_load_values = [(elem * 16) for elem in psutil.getloadavg()]

        self.cpu_load_label0 = QLabel(
            f"CPU load over last 1min:\t {'100.0+' if self.cpu_load_values[0] >= 100.0 else round(self.cpu_load_values[0], 2)}%",
            self,
        )
        self.cpu_group_box_layout.addWidget(self.cpu_load_label0)

        self.cpu_load_label1 = QLabel(
            f"CPU load over last 5mins:\t {'100.0+' if self.cpu_load_values[1] >= 100.0 else round(self.cpu_load_values[1], 2)}%",
            self,
        )
        self.cpu_group_box_layout.addWidget(self.cpu_load_label1)

        self.cpu_load_label2 = QLabel(
            f"CPU load over last 15mins:\t {'100.0+' if self.cpu_load_values[2] >= 100.0 else round(self.cpu_load_values[2], 2)}%",
            self,
        )
        self.cpu_group_box_layout.addWidget(self.cpu_load_label2)
        if self.cpu_load_values[0] >= 30:
            self.cpu_load_label0.setStyleSheet("QLabel {color: red;}")
        elif self.cpu_load_values[0] > 15:
            self.cpu_load_label0.setStyleSheet("QLabel {color: orange;}")
        else:
            self.cpu_load_label0.setStyleSheet("QLabel {color: green;}")
        if self.cpu_load_values[1] >= 30:
            self.cpu_load_label1.setStyleSheet("QLabel {color: red;}")
        elif self.cpu_load_values[1] > 15:
            self.cpu_load_label1.setStyleSheet("QLabel {color: orange;}")
        else:
            self.cpu_load_label1.setStyleSheet("QLabel {color: green;}")
        if self.cpu_load_values[2] >= 30:
            self.cpu_load_label2.setStyleSheet("QLabel {color: red;}")
        elif self.cpu_load_values[2] > 15:
            self.cpu_load_label2.setStyleSheet("QLabel {color: orange;}")
        else:
            self.cpu_load_label2.setStyleSheet("QLabel {color: green;}")

        # Memory summary
        self.memory_group_box = QGroupBox("Memory Summary")
        self.memory_group_box_layout = QVBoxLayout()
        self.memory_group_box_layout.setSpacing(0)
        self.memory_group_box_layout.setAlignment(Qt.AlignTop)
        self.memory_group_box.setLayout(self.memory_group_box_layout)

        self.free_memory_label = QLabel(
            f"Free Memory:\t {human_readable_byte_size(psutil.virtual_memory().available)}, "
            f"({round(100 * psutil.virtual_memory().available / psutil.virtual_memory().total, 2)}%)",
            self,
        )
        self.memory_group_box_layout.addWidget(self.free_memory_label)
        if psutil.virtual_memory().available < 8000000000:
            self.free_memory_label.setStyleSheet("QLabel {color: red;}")
        elif psutil.virtual_memory().available < 32000000000:
            self.free_memory_label.setStyleSheet("QLabel {color: orange;}")
        else:
            self.free_memory_label.setStyleSheet("QLabel {color: green;}")

        self.total_memory_label = QLabel(
            f"Total Memory:\t {human_readable_byte_size(psutil.virtual_memory().total)}",
            self,
        )
        self.memory_group_box_layout.addWidget(self.total_memory_label)
        if psutil.virtual_memory().total < 8000000000:
            self.total_memory_label.setStyleSheet("QLabel {color: red;}")
        elif psutil.virtual_memory().total < 32000000000:
            self.total_memory_label.setStyleSheet("QLabel {color: orange;}")
        else:
            self.total_memory_label.setStyleSheet("QLabel {color: green;}")

        # GPU summary
        self.gpu_group_box = QGroupBox("GPU Summary")
        self.gpu_group_box_layout = QVBoxLayout()
        self.gpu_group_box_layout.setSpacing(0)
        self.gpu_group_box_layout.setAlignment(Qt.AlignTop)
        self.gpu_group_box.setLayout(self.gpu_group_box_layout)

        try:
            cuda_gpu_name = numba.cuda.get_current_device().name.decode()
        except CudaSupportError:
            cuda_gpu_name = "N/A"

        self.cuda_gpu_label = QLabel(f"CUDA GPU: \t\t{cuda_gpu_name}", self)
        self.gpu_group_box_layout.addWidget(self.cuda_gpu_label)

        if cuda_gpu_name != "N/A":
            self.cuda_gpu_label.setStyleSheet("QLabel {color: green;}")
        else:
            self.cuda_gpu_label.setStyleSheet("QLabel {color: red;}")

        cuda_toolkit = numba.cuda.cudadrv.nvvm.is_available()
        self.cudatoolkit_label = QLabel(
            f"CUDA Toolkit: \t\t{'present' if cuda_toolkit else 'absent'}", self
        )
        self.gpu_group_box_layout.addWidget(self.cudatoolkit_label)

        if numba.cuda.cudadrv.nvvm.is_available():
            self.cudatoolkit_label.setStyleSheet("QLabel {color: green;}")
        else:
            self.cudatoolkit_label.setStyleSheet("QLabel {color: red;}")

        try:
            cuda_memory_free = numba.cuda.current_context().get_memory_info().free
            cuda_memory_total = numba.cuda.current_context().get_memory_info().total
        except CudaSupportError:
            cuda_memory_free = 0
            cuda_memory_total = 0

        self.gpu_memory_free_label = QLabel(
            f"Free GPU Memory: \t{human_readable_byte_size(cuda_memory_free)}", self
        )
        self.gpu_group_box_layout.addWidget(self.gpu_memory_free_label)

        self.gpu_memory_total_label = QLabel(
            f"Total GPU Memory: \t{human_readable_byte_size(cuda_memory_total)}", self
        )
        self.gpu_group_box_layout.addWidget(self.gpu_memory_total_label)

        if cuda_memory_total == 0:
            self.gpu_memory_total_label.setStyleSheet("QLabel {color: red;}")
            self.gpu_memory_free_label.setStyleSheet("QLabel {color: red;}")
        else:
            if numba.cuda.current_context().get_memory_info().total < 8000000000:
                self.gpu_memory_total_label.setStyleSheet("QLabel {color: orange;}")
            else:
                self.gpu_memory_total_label.setStyleSheet("QLabel {color: green;}")

            if cuda_memory_free / cuda_memory_total < 0.4:
                self.gpu_memory_free_label.setStyleSheet("QLabel {color: red;}")
            elif cuda_memory_free / cuda_memory_total < 0.8:
                self.gpu_memory_free_label.setStyleSheet("QLabel {color: orange;}")
            else:
                self.gpu_memory_free_label.setStyleSheet("QLabel {color: green;}")

        self.layout.addWidget(self.cpu_group_box)
        self.layout.addWidget(self.memory_group_box)
        self.layout.addWidget(self.gpu_group_box)

        self.setLayout(self.layout)



import numpy
import torch
import torch.nn.functional as F


def richardson_lucy_pytorch(image, psf, iterations=50, clip=True, donut=False):
    use_cuda = True
    device_index = 0
    device = torch.device(f"cuda:{device_index}" if use_cuda else "cpu")
    # print(f"Using device: {device}")

    image = image.astype(numpy.float)
    psf = psf.astype(numpy.float)
    im_deconv = numpy.full(image.shape, image.mean())
    psf_mirror = psf[::-1, ::-1].copy()
    psf_size = psf_mirror.shape[0]

    image = (
        torch.from_numpy(image[numpy.newaxis, numpy.newaxis, ...]).float().to(device)
    )
    psf = torch.from_numpy(psf[numpy.newaxis, numpy.newaxis, ...]).float().to(device)
    psf_mirror = (
        torch.from_numpy(psf_mirror[numpy.newaxis, numpy.newaxis, ...])
        .float()
        .to(device)
    )
    im_deconv = (
        torch.from_numpy(im_deconv[numpy.newaxis, numpy.newaxis, ...])
        .float()
        .to(device)
    )

    for _ in range(iterations):
        pad = (psf_size - 1) // 2
        convolved = F.conv2d(F.pad(im_deconv, (pad, pad, pad, pad), "reflect"), psf)
        relative_blur = image / convolved

        im_deconv *= F.conv2d(
            F.pad(relative_blur, (pad, pad, pad, pad), "reflect"), psf_mirror
        )

    if clip:
        im_deconv[im_deconv > 1] = 1
        im_deconv[im_deconv < 0] = 0

    return im_deconv.detach().cpu().numpy().squeeze()



import numpy
from skimage.exposure import rescale_intensity

from aydin.features.groups.spatial import SpatialFeatures
from aydin.io.datasets import camera


def n(image):
    return rescale_intensity(
        image.astype(numpy.float32), in_range='image', out_range=(0, 1)
    )


def test_spatial_feature_group():
    # get image:
    image = n(camera().astype(numpy.float32))

    # Instantiates spatial features:
    spatial = SpatialFeatures()
    assert spatial.num_features(image.ndim) == image.ndim

    # Check receptive field radius:
    assert spatial.receptive_field_radius == 0

    # Set image:
    spatial.prepare(image)

    # compute features and check their valididty:
    feature_y = numpy.empty_like(image)
    spatial.compute_feature(index=0, feature=feature_y)
    assert feature_y[0, 0] == feature_y[0, 511]
    assert feature_y[0, 0] < feature_y[511, 0]

    feature_x = numpy.empty_like(image)
    spatial.compute_feature(index=1, feature=feature_x)
    assert feature_x[0, 0] < feature_x[0, 511]
    assert feature_x[0, 0] == feature_x[511, 0]



import numpy


def outer_product(*arrays):

    outer = numpy.ix_(*arrays)
    shapes = tuple(m.shape for m in outer)
    shape = numpy.max(numpy.array(shapes), axis=0)

    result = numpy.ones(shape, dtype=arrays[0].dtype)
    for coordinate_map in outer:
        result *= numpy.array(coordinate_map)
    return result


def outer_sum(*arrays):

    outer = numpy.ix_(*arrays)
    shapes = tuple(m.shape for m in outer)
    shape = numpy.max(numpy.array(shapes), axis=0)

    result = numpy.zeros(shape, dtype=arrays[0].dtype)
    for coordinate_map in outer:
        result += numpy.array(coordinate_map)
    return result



# flake8: noqa
from aydin.io.datasets import cropped_newyork
from aydin.util.j_invariance.demo.demo_j_invariant_smart import demo_j_invariant_smart


def test_j_invariant_smart():
    demo_j_invariant_smart(cropped_newyork(crop_amount=470), display=False)



import os
import tempfile
from os import makedirs
from os.path import join, exists
from sys import platform


def get_home_folder():
    from pathlib import Path

    home_folder = f"{Path.home()}"
    return home_folder


def get_temp_folder():

    temp_folder = None

    if platform == "linux" or platform == "linux2":
        temp_folder = tempfile.gettempdir()

    elif platform == "darwin":
        temp_folder = tempfile.gettempdir()

    elif platform == "win32":
        temp_folder = tempfile.gettempdir()

    try:
        makedirs(temp_folder)
    except Exception:
        pass

    if exists(temp_folder):
        return temp_folder
    else:
        return None


def get_cache_folder():

    cache_folder = None

    if platform == "linux" or platform == "linux2":
        cache_folder = join(get_home_folder(), '.cache')

    elif platform == "darwin":
        cache_folder = join(get_home_folder(), '/Library/Caches')

    elif platform == "win32":
        cache_folder = join(get_home_folder(), os.getenv('LOCALAPPDATA'))

    try:
        makedirs(cache_folder)
    except Exception:
        pass

    if exists(cache_folder):
        return cache_folder
    else:
        return None



# Function to add camera noise
from typing import Optional

import numpy
from numpy.random import RandomState


# from numpy.typing import ArrayLike


def simulate_camera_image(
    photons_per_second,
    exposure_time_s: float = 0.20,
    quantum_efficiency: float = 0.82,
    gain: float = 2.22,
    gain_sigma: float = 0.02,
    gain_column_sigma: float = 0.01,
    offset_mean: float = 1.63,
    offset_sigma: float = 1,
    dark_current: float = 0.04,
    dark_current_sigma: float = 0.01,
    dark_current_column_sigma: float = 0.01,
    min_exposure_dark_current: float = 0.001,
    num_hot_pixels: int = 8,
    num_cold_pixels: int = 8,
    probability_cosmic_ray: float = 1e-6,
    bitdepth: int = 12,
    baseline: int = 100,
    shot_rnd: Optional[RandomState] = None,
    camera_rnd: Optional[RandomState] = None,
    dtype=numpy.int32,
):
    """
    Realistic noise simulation for scientific cameras.
    Adapted from Kyle Douglas blog: http://kmdouglass.github.io/posts/modeling-noise-for-image-simulations/
    With additional ideas from: https://mwcraig.github.io/ccd-as-book/01-03-Construction-of-an-artificial-but-realistic-image.html


    Parameters
    ----------
    photons_per_second :
        Image representing the number of photons received on the camera per pixel per second
    exposure_time_s :
        Exposure time in seconds
    quantum_efficiency :
        Quantum efficiency - i.e. conversion factor between photons and electrons
    gain :
        Conversion factor between electrons and Analog-Digital-Units (ADU)
    gain_sigma :
        Unfortunately, not all gains are identical across the camera pixels,
        This parameter controls the spread of the gain.
    gain_column_sigma :
        And often each column of the detector has its own electronics that induce another source of column-dependent noise.
        This parameter controls the additional spread of the gain per column.
    offset_mean :
        Pixel  amplification offset noise mean value.
    offset_sigma :
        Pixel amplification offset noise sigma value.
    dark_current :
        Dark current, in electrons per pixel per second, which is the way manufacturers typically
        report it.
    dark_current_sigma :
        Unfortunately, the dark current is not identical for each and every pixel.
        This parameter controls the spread of the dark current.
    dark_current_column_sigma :
        And often each column of the detector has its own electronics that induce another source of column-dependent noise.
        This parameter controls the additional spread of the dark current per column.
    min_exposure_dark_current:
        Minimal exposure for the purpose of dark photons. The effects of the dark current do not completely vanish for very short exposures...
    num_hot_pixels:
        Number of hot pixels.
    num_cold_pixels:
        Number of cold pixels.
    probability_cosmic_ray:
        Probability per pixel per second that a cosmic ray will hit a camera pixel.
    bitdepth :
        Bit depth of each pixel fo the camera
    baseline :
        Baseline value for camera
    shot_rnd :
        Random state for each image (time dependent)
    camera_rnd :
            Random state for each camera (time indedependent, camera instance dependent)
    dtype :
        Integral dtype to return image in

    Returns
    -------

    """

    if shot_rnd is None:
        shot_rnd = numpy.random.RandomState(seed=None)

    if camera_rnd is None:
        camera_rnd = numpy.random.RandomState(seed=42)

    # Gain image:
    gain_image = gain * numpy.ones_like(photons_per_second, dtype=numpy.float32)

    # Unfortunately the gain is a bit different for each pixel:
    if gain_sigma > 0:
        gain_image += _normal(
            camera_rnd, scale=gain_sigma, size=photons_per_second.shape
        )

    # And often each column of the detector has its own electronics that induce another source of column-dependent noise:
    if gain_column_sigma > 0:
        gain_image += _normal(
            camera_rnd, scale=gain_column_sigma, size=photons_per_second.shape[1:]
        )[numpy.newaxis, :]

    # Clip gain:
    gain_image = numpy.clip(gain_image, a_min=0, a_max=None, out=gain_image)

    # Readout offset noise:
    offset_image = offset_mean + _normal(
        shot_rnd, scale=offset_sigma, size=photons_per_second.shape
    )

    # Dark current image:
    dark_current_image = dark_current * numpy.ones_like(
        photons_per_second, dtype=numpy.float32
    )

    # Unfortunately the dark current is a bit different for each pixel:
    if dark_current_sigma > 0:
        dark_current_image += _normal(
            camera_rnd, scale=dark_current_sigma, size=photons_per_second.shape
        )

    # And often each column of the detector has its own electronics that induce another source of column-dependent noise:
    if dark_current_column_sigma > 0:
        dark_current_image += _normal(
            camera_rnd,
            scale=dark_current_column_sigma,
            size=photons_per_second.shape[1:],
        )[numpy.newaxis, :]

    # Add shot noise
    photons = _poisson(
        shot_rnd, photons_per_second * exposure_time_s, size=photons_per_second.shape
    )

    # Converts from photons to electrons:
    electrons = quantum_efficiency * photons

    # Epsilon value for clipping lowgain:
    epsilon = 1e-6

    # dark current electrons:
    dark_electrons = _poisson(
        shot_rnd,
        numpy.clip(dark_current_image, a_min=epsilon, a_max=None)
        * max(min_exposure_dark_current, exposure_time_s),
        size=photons.shape,
    )

    # Cosmic rays (lol):
    if probability_cosmic_ray > 0:
        num_of_rays = exposure_time_s * probability_cosmic_ray * electrons.size
        effective_num_of_rays = shot_rnd.poisson(num_of_rays)

        y_max, x_max = electrons.shape
        ray_x = camera_rnd.randint(0, x_max, size=effective_num_of_rays)
        ray_y = camera_rnd.randint(0, y_max, size=effective_num_of_rays)
        dark_electrons[tuple([ray_y, ray_x])] += int(gain * 16)

    # Some pixels are hot:
    if num_hot_pixels > 0:
        y_max, x_max = dark_electrons.shape
        hot_x = camera_rnd.randint(0, x_max, size=num_hot_pixels)
        hot_y = camera_rnd.randint(0, y_max, size=num_hot_pixels)
        dark_electrons[tuple([hot_y, hot_x])] *= min(16, 2 ** (bitdepth - 2))

    # Some pixels are cold:
    if num_cold_pixels > 0:
        y_max, x_max = electrons.shape
        cold_x = camera_rnd.randint(0, x_max, size=num_cold_pixels)
        cold_y = camera_rnd.randint(0, y_max, size=num_cold_pixels)
        electrons[tuple([cold_y, cold_x])] /= min(16, 2 ** (bitdepth - 2))

    # Add dark current
    all_electrons = dark_electrons + electrons

    # max ADU:
    max_adu = numpy.int(2 ** bitdepth - 1)

    # Convert to discrete numbers (ADU):
    adu = (all_electrons * gain_image + offset_image).astype(dtype)

    # Add baseline:
    adu += baseline

    # Models pixel saturation:
    adu[adu > max_adu] = max_adu

    return adu


def _poisson(rnd: RandomState, lam, size):
    return rnd.poisson(lam=lam, size=size)


def _normal(rnd: RandomState, scale, size):
    return rnd.normal(scale=scale, size=size)



import math
from functools import partial
from math import prod
from typing import Optional, Union, Tuple
import numpy
from sklearn.decomposition import PCA

from aydin.util.crop.rep_crop import representative_crop
from aydin.util.j_invariance.j_invariant_smart import calibrate_denoiser_smart
from aydin.util.patch_size.patch_size import default_patch_size
from aydin.util.patch_transform.patch_transform import (
    extract_patches_nd,
    reconstruct_from_nd_patches,
)


def calibrate_denoise_pca(
    image,
    patch_size: Optional[Union[int, Tuple[int], str]] = None,
    crop_size_in_voxels: Optional[int] = None,
    max_num_evaluations: int = 20,
    display_images: bool = False,
    **other_fixed_parameters,
):
    """
    Calibrates the Principal Component Analysis (PCA) denoiser for the given
    image and returns the optimal parameters obtained using the N2S loss.

    Parameters
    ----------
    image: ArrayLike
        Image to calibrate spectral denoiser for.

    patch_size: int
        Patch size for the 'image-to-patch' transform.
        Can be an int s that corresponds to isotropic patches of shape: (s,)*image.ndim,
        or a tuple of ints. By default (None) the patch size is chosen automatically
        to give the best results.
        (advanced)

    crop_size_in_voxels: int or None for default
        Number of voxels for crop used to calibrate denoiser.
        (advanced)

    max_num_evaluations: int
        Maximum number of evaluations for finding the optimal parameters.
        (advanced)

    display_images: bool
        When True the denoised images encountered during optimisation are shown

    other_fixed_parameters: dict
        Any other fixed parameters

    Returns
    -------
    Denoising function, dictionary containing optimal parameters,
    and free memory needed in bytes for computation.
    """

    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # obtain representative crop, to speed things up...
    crop = representative_crop(image, crop_size=crop_size_in_voxels, display_crop=False)

    # Normalise patch size:
    patch_size = default_patch_size(image, patch_size, odd=True)

    # Ranges:
    threshold_range = (0.0, 1.0)

    # Parameters to test when calibrating the denoising algorithm
    parameter_ranges = {'threshold': threshold_range}

    # Combine fixed parameters:
    other_fixed_parameters = other_fixed_parameters | {'patch_size': patch_size}

    # Partial function:
    _denoise_pca = partial(
        denoise_pca, **(other_fixed_parameters | {'multi_core': False})
    )

    # Calibrate denoiser
    best_parameters = (
        calibrate_denoiser_smart(
            crop,
            _denoise_pca,
            denoise_parameters=parameter_ranges,
            display_images=display_images,
            max_num_evaluations=max_num_evaluations,
        )
        | other_fixed_parameters
    )

    # Memory needed:
    memory_needed = 2 * image.nbytes + 6 * image.nbytes * math.prod(patch_size)

    return denoise_pca, best_parameters, memory_needed


def denoise_pca(
    image,
    patch_size: Optional[Union[int, Tuple[int]]] = None,
    threshold: float = 0.1,
    reconstruction_gamma: float = 0,
    multi_core: bool = True,
):
    """
    Denoises the given image by first applying the patch
    transform, and then doing a PCA projection of these patches
    along the first components. The cut-off is set by a
    threshold parameter.

    Parameters
    ----------
    image: ArrayLike
        Image to denoise

    patch_size: int
        Patch size for the 'image-to-patch' transform.
        Can be: 'full' for a single patch covering the whole image, 'half', 'quarter',
        or an int s that corresponds to isotropic patches of shape: (s,)*image.ndim,
        or a tuple of ints. By default (None) the patch size is chosen automatically
        to give the best results.

    threshold: float
        Threshold for proportion of components to retain. Between 0 and 1

    reconstruction_gamma: float
        Patch reconstruction parameter

    multi_core: bool
        By default we use as many cores as possible, in some cases, for small
        (test) images, it might be faster to run on a single core instead of
        starting the whole parallelization machinery.


    Returns
    -------
    Denoised image

    """

    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # Normalise patch size:
    patch_size = default_patch_size(image, patch_size, odd=True)

    # First we apply the patch transform:
    patches = extract_patches_nd(image, patch_size=patch_size)

    # reshape patches as vectors:
    original_patches_shape = patches.shape
    patches = patches.reshape(patches.shape[0], -1)

    # PCA dim reduction setup:
    n_components = 1 + max(0, int(threshold * (prod(patch_size) - 1)))
    pca = PCA(n_components=n_components)

    # Project patches:
    patches = pca.inverse_transform(pca.fit_transform(patches))

    # reshape patches back to their original shape:
    patches = patches.reshape(original_patches_shape)

    # Transform back from patches to image:
    denoised_image = reconstruct_from_nd_patches(
        patches, image.shape, gamma=reconstruction_gamma
    )

    # Cast back to float32 if needed:
    denoised_image = denoised_image.astype(numpy.float32, copy=False)

    return denoised_image



import sys
from qtpy import QtCore


class OutputWrapper(QtCore.QObject):
    outputWritten = QtCore.Signal(object, object)

    def __init__(self, parent, stdout=True):
        QtCore.QObject.__init__(self, parent)
        if stdout:
            self._stream = sys.stdout
            sys.stdout = self
        else:
            self._stream = sys.stderr
            sys.stderr = self
        self._stdout = stdout

    def write(self, text):
        self._stream.write(text)
        self.outputWritten.emit(text, self._stdout)

    def __getattr__(self, name):
        return getattr(self._stream, name)

    def __del__(self):
        try:
            if self._stdout:
                sys.stdout = self._stream
            else:
                sys.stderr = self._stream
        except AttributeError:
            pass



from functools import partial
from typing import Optional
import numpy
import numpy as np
from skimage.restoration import denoise_bilateral as skimage_denoise_bilateral

from aydin.util.crop.rep_crop import representative_crop
from aydin.util.denoise_nd.denoise_nd import extend_nd
from aydin.util.j_invariance.j_invariant_classic import calibrate_denoiser_classic


def calibrate_denoise_bilateral(
    image,
    bins: int = 10000,
    crop_size_in_voxels: Optional[int] = None,
    display_images: bool = False,
    **other_fixed_parameters,
):
    """
    Calibrates the bilateral denoiser for the given image and returns the optimal
    parameters obtained using the N2S loss.

    Note: it seems that the bilateral filter of scikit-image
    is broken!

    Parameters
    ----------
    image: ArrayLike
        Image to calibrate denoiser for.

    bins: int
        Number of discrete values for Gaussian weights of
        color filtering. A larger value results in improved
        accuracy.

    crop_size_in_voxels: int or None for default
        Number of voxels for crop used to calibrate
        denoiser.

    display_images: bool
        When True the denoised images encountered during
        optimisation are shown

    other_fixed_parameters: dict
        Any other fixed parameters

    Returns
    -------
    Denoising function, dictionary containing optimal parameters,
    and free memory needed in bytes for computation.
    """
    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    # obtain representative crop, to speed things up...
    crop = representative_crop(image, crop_size=crop_size_in_voxels)

    # Sigma spatial range:
    sigma_spatial_range = np.arange(0.01, 1, 0.05) ** 1.5

    # Sigma color range:
    sigma_color_range = np.arange(0.01, 1, 0.05) ** 1.5

    # Parameters to test when calibrating the denoising algorithm
    parameter_ranges = {
        'sigma_spatial': sigma_spatial_range,
        'sigma_color': sigma_color_range,
    }

    # Combine fixed parameters:
    other_fixed_parameters = other_fixed_parameters | {'bins': bins}

    # Partial function:
    _denoise_bilateral = partial(denoise_bilateral, **other_fixed_parameters)

    # Calibrate denoiser
    best_parameters = (
        calibrate_denoiser_classic(
            crop,
            _denoise_bilateral,
            denoise_parameters=parameter_ranges,
            display_images=display_images,
        )
        | other_fixed_parameters
    )

    # Memory needed:
    memory_needed = 2 * image.nbytes

    return denoise_bilateral, best_parameters, memory_needed


def denoise_bilateral(image, sigma_color=None, sigma_spatial=1, bins=10000, **kwargs):
    """
    Denoises the given image using a <a
    href="https://en.wikipedia.org/wiki/Bilateral_filter">bilateral
    filter</a>.
    The bilateral filter is a edge-preserving smoothing filter that can
    be used for image denoising. Each pixel value is replaced by a
    weighted average of intensity values from nearby pixels. The
    weighting is inversely related to the pixel distance in space but
    also in the pixels value differences.

    Parameters
    ----------
    image : ArrayLike
        Image to denoise

    sigma_color : float
        Standard deviation for grayvalue/color distance (radiometric
        similarity). A larger value results in averaging of pixels with larger
        radiometric differences. Note, that the image will be converted using
        the `img_as_float` function and thus the standard deviation is in
        respect to the range ``[0, 1]``. If the value is ``None`` the standard
        deviation of the ``image`` will be used.

    sigma_spatial : float
        Standard deviation for range distance. A larger value results in
        averaging of pixels with larger spatial differences.

    bins : int
        Number of discrete values for Gaussian weights of color filtering.
        A larger value results in improved accuracy.

    kwargs: dict
        Other parameters

    Returns
    -------
    Denoised image

    """

    # Convert image to float if needed:
    image = image.astype(dtype=numpy.float32, copy=False)

    _skimage_denoise_bilateral = extend_nd(available_dims=[2])(
        skimage_denoise_bilateral
    )

    return _skimage_denoise_bilateral(
        image,
        sigma_color=sigma_color,
        sigma_spatial=sigma_spatial,
        bins=bins,
        mode='reflect',
        **kwargs,
    )



import numpy
from skimage.data import binary_blobs

from aydin.io.datasets import normalise
from aydin.it.transforms.attenuation import AttenuationTransform


def test_attenuation_correction_real():
    image = binary_blobs(length=128, seed=1, n_dim=3).astype(numpy.float32)
    image = normalise(image)

    ramp = numpy.linspace(0, 1, 128).astype(numpy.float32)

    attenuated = image * ramp

    ac = AttenuationTransform(axes=2)

    preprocessed = ac.preprocess(attenuated)
    postprocessed = ac.postprocess(preprocessed)

    # import napari
    # with napari.gui_qt():
    #     viewer = napari.Viewer()
    #     viewer.add_image(image, name='image')
    #     viewer.add_image(attenuated, name='attenuated')
    #     viewer.add_image(preprocessed, name='preprocessed')
    #     viewer.add_image(postprocessed, name='postprocessed')

    assert image.shape == postprocessed.shape
    assert image.dtype == postprocessed.dtype
    assert numpy.abs(preprocessed - image).mean() < 0.005

    assert postprocessed.dtype == attenuated.dtype
    assert numpy.abs(postprocessed - attenuated).mean() < 1e-8



from math import ceil

import numba
import numpy
from numba import jit, prange

__fastmath = {'contract', 'afn', 'reassoc'}
__error_model = 'numpy'


def numba_cpu_uniform_filter(
    image, size=3, output=None, mode="nearest", cval=0.0, origin=0
):
    # Save original image dtype:
    original_dtype = image.dtype

    # Numba does not support float16 yet:
    dtype = numpy.float32 if original_dtype == numpy.float16 else original_dtype
    image = image.astype(dtype=dtype, copy=False)

    # Instantiates working images:
    image_a_0 = numpy.empty(image.shape, dtype=dtype)
    image_b_0 = numpy.empty(image.shape, dtype=dtype)

    # Current output (just a ref)
    output = None

    axes = list(range(image.ndim))
    if len(axes) == 1:
        _cpu_line_filter(image, image_a_0, size)
        output = image_a_0
    elif len(axes) > 0:
        for axis in axes:

            if axis == 0:
                image_a = image
                image_b = image_a_0

            filter_size = size[axis] if isinstance(size, tuple) else size

            # lprint(f"axis: {axis}, filter_size: {filter_size}")

            # set the parallelism:
            parallelism = numba.get_num_threads()
            # lprint(f"Number of threads: {parallelism}")
            # max(1, int(0.9*multiprocessing.cpu_count()))

            uniform_filter1d_with_conditionals(
                image_a, image_b, filter_size, axis, parallelism=parallelism
            )

            output = image_b

            if axis == 0:
                image_a = image_a_0
                image_b = image_b_0
            else:
                image_b, image_a = image_a, image_b
    else:
        output = image.copy()

    # Make sure that output image has correct dtype:
    output = output.astype(dtype=original_dtype, copy=False)

    return output


def uniform_filter1d_with_conditionals(image, output, filter_size, axis, parallelism=8):
    if image.ndim == 1:
        _cpu_line_uniform_filter_without_loops(image, output, filter_size, parallelism)
    elif image.ndim == 2:
        prepared_image = image.swapaxes(0, 1) if axis == 0 else image
        prepared_output = output.swapaxes(0, 1) if axis == 0 else output
        _cpu_line_uniform_filter_without_loops(
            prepared_image, prepared_output, filter_size, parallelism
        )
    elif image.ndim == 3:
        prepared_image = image.swapaxes(axis, 2) if axis != 2 else image
        prepared_output = output.swapaxes(axis, 2) if axis != 2 else output
        _cpu_line_uniform_filter_with_2d_loop(
            prepared_image, prepared_output, filter_size, parallelism
        )
        # cpu_line_uniform_filter_with_2d_loop.parallel_diagnostics(level=4)
        # for key, value in cpu_line_filter.inspect_asm().items():
        #     print(f"{key} -> {value}")

    elif image.ndim == 4:
        prepared_image = image.swapaxes(axis, 3) if axis != 3 else image
        prepared_output = output.swapaxes(axis, 3) if axis != 3 else output
        _cpu_line_uniform_filter_with_3d_loop(
            prepared_image, prepared_output, filter_size, parallelism
        )
        # cpu_line_uniform_filter_with_3d_loop.parallel_diagnostics(level=4)


@jit(nopython=True, parallel=True, error_model=__error_model, fastmath=__fastmath)
def _cpu_line_uniform_filter_without_loops(image, output, filter_size, parallelism=8):
    """
    Numba jitted and parallelized method to apply uniform filter across
    last axis of the image provided. Doesn't return anything, output array
    should be provided as an argument.

    Parameters
    ----------
    image
    output
    filter_size
    parallelism

    """

    length = image.shape[0]
    chunk_length = int(ceil(length / parallelism))

    for c in prange(parallelism):
        for k in range(chunk_length):
            i = k + c * chunk_length
            if i < length:
                input_line = image[i, :]
                output_line = output[i, :]
                _cpu_line_filter(input_line, output_line, filter_size)

    # print(cpu_line_filter.inspect_llvm())


@jit(nopython=True, parallel=True, error_model=__error_model, fastmath=__fastmath)
def _cpu_line_uniform_filter_with_2d_loop(image, output, filter_size, parallelism=8):
    """
    Numba jitted and parallelized method to apply uniform filter across
    last axis of the image provided. Doesn't return anything, output array
    should be provided as an argument.

    Parameters
    ----------
    image
    output
    filter_size

    """

    height = image.shape[0]
    width = image.shape[1]
    chunk_height = int(ceil(height / parallelism))

    for c in prange(parallelism):
        for k in range(chunk_height):
            y = k + c * chunk_height
            image_y = image[y]
            output_y = output[y]
            if y < height:
                for x in range(width):
                    input_line = image_y[x]
                    output_line = output_y[x]
                    _cpu_line_filter(input_line, output_line, filter_size)


@jit(nopython=True, parallel=True, error_model=__error_model, fastmath=__fastmath)
def _cpu_line_uniform_filter_with_3d_loop(image, output, filter_size, parallelism=8):
    """
    Numba jitted and parallelized method to apply uniform filter across
    last axis of the image provided. Doesn't return anything, output array
    should be provided as an argument.

    Parameters
    ----------
    image
    output
    filter_size

    """

    depth = image.shape[0]
    height = image.shape[1]
    width = image.shape[2]
    chunk_depth = int(ceil(depth / parallelism))

    for c in prange(parallelism):
        for k in range(chunk_depth):
            z = k + c * chunk_depth
            if z < depth:
                image_z = image[z]
                output_z = output[z]
                for y in range(height):
                    image_y = image_z[y]
                    output_y = output_z[y]
                    for x in range(width):
                        input_line = image_y[x]
                        output_line = output_y[x]
                        _cpu_line_filter(input_line, output_line, filter_size)


@jit(nopython=True, error_model=__error_model, fastmath=__fastmath)
def _cpu_line_filter(input_line, output_line, filter_size):
    """
    Numba jitted line filter implementation. Doesn't return anything,
    output array should be provided as an argument.

    Parameters
    ----------
    input_line
        1D input array
    output_line
        1D output array
    filter_size
        Size of the uniform filter
    """

    def safe_index(index, size):
        if index < 0:
            return 0
        elif index >= size:
            return size - 1
        else:
            return index

    array_size = len(input_line)
    left_offset, right_offset = (filter_size // 2, filter_size - (filter_size // 2) - 1)

    tmp = numpy.float64(0.0)

    for ind in range(-left_offset, right_offset + 1):
        tmp += input_line[safe_index(ind, array_size)]

    output_line[0] = tmp / filter_size

    for ind in range(1, array_size):
        element_to_add_index = safe_index(ind + right_offset, array_size)
        element_to_sub_index = safe_index(ind - left_offset - 1, array_size)

        tmp += input_line[element_to_add_index]
        tmp -= input_line[element_to_sub_index]
        output_line[ind] = tmp / filter_size



# flake8: noqa
import time

import numpy
import numpy as np
from numpy.fft import fftshift
from skimage.data import camera
from skimage.metrics import peak_signal_noise_ratio as psnr
from skimage.metrics import structural_similarity as ssim

from aydin.analysis.find_kernel import compute_relative_blur_kernel
from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import (
    newyork,
    pollen,
    normalise,
    add_noise,
    lizard,
    dots,
    characters,
)
from aydin.it.deconvolution.lr_deconv import ImageTranslatorLRDeconv
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.lgbm import LGBMRegressor
from aydin.util.log.log import Log


def demo(image, name):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    Log.enable_output = True
    # Log.set_log_max_depth(5)

    image = normalise(image.astype(np.float32))
    noisy = add_noise(image)

    generator = StandardFeatureGenerator(
        include_corner_features=True,
        include_scale_one=False,
        include_fine_features=True,
        include_line_features=True,
        include_spatial_features=True,
    )
    regressor = LGBMRegressor(patience=16, compute_training_loss=True)

    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)

    start = time.time()
    it.train(noisy, noisy)
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")

    # in case of batching we have to do this:
    start = time.time()
    denoised = it.translate(noisy)
    stop = time.time()
    print(f"inference: elapsed time:  {stop - start} ")

    computed_kernel = numpy.log1p(
        numpy.abs((fftshift(compute_relative_blur_kernel(denoised, image))))
    )

    computed_kernel = computed_kernel[256 - 32 : 256 + 31 + 1, 256 - 32 : 256 + 32 + 1]

    psf_kernel = numpy.asarray([[0.5, 1, 0.5], [1, 1, 1], [0.5, 1, 0.5]])
    psf_kernel = psf_kernel / psf_kernel.sum()
    lr = ImageTranslatorLRDeconv(
        psf_kernel=psf_kernel, max_num_iterations=2, backend='cupy'
    )
    lr.train(denoised, denoised)
    denoised_deconv = lr.translate(denoised)

    image = numpy.clip(image, 0, 1)
    noisy = numpy.clip(noisy, 0, 1)
    denoised = numpy.clip(denoised, 0, 1)
    denoised_deconv = numpy.clip(denoised_deconv, 0, 1)

    psnr_noisy = psnr(image, noisy)
    ssim_noisy = ssim(image, noisy)

    psnr_denoised = psnr(image, denoised)
    ssim_denoised = ssim(image, denoised)

    psnr_denoised_deconv = psnr(image, denoised_deconv)
    ssim_denoised_deconv = ssim(image, denoised_deconv)

    print("noisy          :", psnr_noisy, ssim_noisy)
    print("denoised       :", psnr_denoised, ssim_denoised)
    print("denoised_deconv:", psnr_denoised_deconv, ssim_denoised_deconv)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(normalise(computed_kernel), name='computed_kernel')
        viewer.add_image(normalise(image), name='image')
        viewer.add_image(normalise(noisy), name='noisy')
        viewer.add_image(normalise(denoised), name='denoised')
        viewer.add_image(normalise(denoised_deconv), name='denoised_deconv')


characters_image = characters()
demo(characters_image, "characters")
camera_image = camera()
demo(camera_image, "camera")
lizard_image = lizard()
demo(lizard_image, "lizard")
pollen_image = pollen()
demo(pollen_image, "pollen")
dots_image = dots()
demo(dots_image, "dots")
newyork_image = newyork()
demo(newyork_image, "newyork")



# flake8: noqa
from aydin.util.string.break_text import break_text

_test_line_breaks_text = """    Suppresses fixed, axis aligned, offset patterns along any combination of
    axis. Given a list of lists of axis that defines axis-aligned volumes,
    intensity fluctuations of these volumes are stabilised. You can suppress
    intensity fluctuation over time, suppress fixed offsets per pixel over
    time, suppress intensity fluctuations per row, per column, and more...

    For example, assume an image with dimensions tyx (t+2D), and you want to
    suppress fluctuations of intensity along the t axis, then you provide:
    axes=[[0]] (or simply 0 or [0]) which means that the average intensity
    for all planes along t (axis=0) will be stabilised. If instead you want
    to suppress some fixed background offset over xy planes, then you do:
    axes=[[1,2]]. If you want to do both, then you use: axes=[[0], [1,
    2]]. Please note that these corrections are applied in the order
    specified by the list of axis combinations. It is not recommended to
    reapply the pattern after denoising, unless the pattern itself is of
    value and is not considered noise.
    
        For images with little noise, applying a high-pass filter can help denoise the image by removing some of the
    image complexity. The low-frequency parts of the image do not need to be denoised because sometimes the challenge
    is disentangling the (high-frequency) noise from the high-frequencies in the image. The scale parameter must be
    chosen with care. The lesser the noise, the smaller the value. Values around 1 work well but must be tuned
    depending on the image. If the scale parameter is too low, some noise might be left untouched. The best is to
    keep the parameter as low as possible while still achieving good denoising performance. It is also possible to
    apply median filtering when computing the low-pass image which helps reducing the impact of outlier voxel values,
    for example salt&pepper noise. Note: when median filtering is on, larger values of sigma (e.g. >= 1) are
    recommended, unless when the level of noise is very low in which case a sigma of 0 (no Gaussian blur) may be
    advantageous. To recover the original denoised image the filtering is undone during post-processing. Note: this
    is ideal for treating <a href='https://en.wikipedia.org/wiki/Colors_of_noise'>'blue' noise</a> that is
    characterised by a high-frequency support."""


def test_line_breaks():

    print("")
    print(break_text(_test_line_breaks_text, 80))



import json
import os

from aydin.util.log.log import lprint


def absPath(myPath):
    """Get absolute path to resource, works for dev and for PyInstaller"""
    import sys

    try:
        # PyInstaller creates a temp folder and stores path in _MEIPASS
        base_path = sys._MEIPASS
        lprint("found MEIPASS: %s " % os.path.join(base_path, os.path.basename(myPath)))

        return os.path.join(base_path, os.path.basename(myPath))
    except Exception as e:
        lprint("did not find MEIPASS: %s " % e)

        base_path = os.path.abspath(os.path.dirname(__file__))
        return os.path.join(base_path, myPath)


class JSONResourceLoader:
    def __init__(self, resource_file_name):
        if resource_file_name is None or resource_file_name == "":
            raise ValueError(
                "JSONResourceLoader has to be initiated with a resource file name argument."
            )

        with open(absPath(resource_file_name)) as json_file:
            self.json = json.load(json_file)



import numpy
from numpy.linalg import norm
from scipy.fft import dct
from skimage.metrics import mean_squared_error


def spectral_psnr(norm_true_image, norm_test_image):
    """Spectral PSNR calculation

    Parameters
    ----------
    norm_true_image : numpy.typing.ArrayLike
    norm_test_image : numpy.typing.ArrayLike

    Returns
    -------
    Calculated PSNR : float

    Notes
    -----
    Interesting package: https://github.com/andrewekhalel/sewar
    """
    norm_true_image = norm_true_image / norm(norm_true_image.flatten(), 2)
    norm_test_image = norm_test_image / norm(norm_test_image.flatten(), 2)

    dct_norm_true_image = dct(
        dct(norm_true_image, axis=0, workers=-1), axis=1, workers=-1
    )
    dct_norm_test_image = dct(
        dct(norm_test_image, axis=0, workers=-1), axis=1, workers=-1
    )

    norm_dct_norm_true_image = dct_norm_true_image / norm(
        dct_norm_true_image.flatten(), 2
    )
    norm_dct_norm_test_image = dct_norm_test_image / norm(
        dct_norm_test_image.flatten(), 2
    )

    norm_true_image = abs(norm_dct_norm_true_image)
    norm_test_image = abs(norm_dct_norm_test_image)

    err = mean_squared_error(norm_true_image, norm_test_image)
    return 10 * numpy.log10(1 / err)


def spectral_mutual_information(image_a, image_b, normalised: bool = True):
    """Spectral mutual information

    Parameters
    ----------
    image_a : numpy.typing.ArrayLike
    image_b : numpy.typing.ArrayLike
    normalised : bool

    Returns
    -------
    mutual_information

    """
    norm_image_a = image_a / norm(image_a.flatten(), 2)
    norm_image_b = image_b / norm(image_b.flatten(), 2)

    dct_norm_true_image = dct(dct(norm_image_a, axis=0, workers=-1), axis=1, workers=-1)
    dct_norm_test_image = dct(dct(norm_image_b, axis=0, workers=-1), axis=1, workers=-1)

    return mutual_information(
        dct_norm_true_image, dct_norm_test_image, normalised=normalised
    )


def joint_information(image_a, image_b, bins: int = 256):
    """Joint information

    Parameters
    ----------
    image_a : numpy.typing.ArrayLike
    image_b : numpy.typing.ArrayLike
    bins : int

    Returns
    -------
    joint information

    """
    image_a = image_a.flatten()
    image_b = image_b.flatten()

    c_xy = numpy.histogram2d(image_a, image_b, bins)[0]
    ji = joint_entropy_from_contingency(c_xy)
    return ji


def mutual_information(image_a, image_b, bins: int = 256, normalised: bool = True):
    """Mutual information

    Parameters
    ----------
    image_a : numpy.typing.ArrayLike
    image_b : numpy.typing.ArrayLike
    bins : int
    normalised : bool

    Returns
    -------
    mutual information

    """
    image_a = image_a.flatten()
    image_b = image_b.flatten()

    c_xy = numpy.histogram2d(image_a, image_b, bins)[0]
    mi = mutual_info_from_contingency(c_xy)
    mi = mi / joint_entropy_from_contingency(c_xy) if normalised else mi
    return mi


def joint_entropy_from_contingency(contingency):
    """Joint entropy from contingency

    Parameters
    ----------
    contingency : numpy.typing.ArrayLike

    Returns
    -------
    Joint entropy from contingency

    """

    # cordinates of non-zero entries in contingency table:
    nzx, nzy = numpy.nonzero(contingency)

    # non zero values:
    nz_val = contingency[nzx, nzy]

    # sum of all values in contingnecy table:
    contingency_sum = contingency.sum()

    # normalised contingency, i.e. probability:
    p = nz_val / contingency_sum

    # log contingency:
    log_p = numpy.log2(p)

    # Joint entropy:
    joint_entropy = -p * log_p

    return joint_entropy.sum()


def mutual_info_from_contingency(contingency):
    """Mutual info from contingency

    Parameters
    ----------
    contingency : numpy.typing.ArrayLike

    Returns
    -------
    Mutual info from contingency

    """

    # cordinates of non-zero entries in contingency table:
    nzx, nzy = numpy.nonzero(contingency)

    # non zero values:
    nz_val = contingency[nzx, nzy]

    # sum of all values in contingnecy table:
    contingency_sum = contingency.sum()

    # marginals:
    pi = numpy.ravel(contingency.sum(axis=1))
    pj = numpy.ravel(contingency.sum(axis=0))

    #
    log_contingency_nm = numpy.log2(nz_val)
    contingency_nm = nz_val / contingency_sum
    # Don't need to calculate the full outer product, just for non-zeroes
    outer = pi.take(nzx).astype(numpy.int64, copy=False) * pj.take(nzy).astype(
        numpy.int64, copy=False
    )
    log_outer = -numpy.log2(outer) + numpy.log2(pi.sum()) + numpy.log2(pj.sum())
    mi = (
        contingency_nm * (log_contingency_nm - numpy.log2(contingency_sum))
        + contingency_nm * log_outer
    )
    return mi.sum()



import torch.nn as nn


class DnCNN(nn.Module):
    def __init__(
        self,
        n_channel_in=1,
        n_channel_out=1,
        num_of_layers=17,
        kernel_size=3,
        padding=1,
        features=64,
    ):
        super(DnCNN, self).__init__()

        layers = []
        layers.append(
            nn.Conv2d(
                in_channels=n_channel_in,
                out_channels=features,
                kernel_size=kernel_size,
                padding=padding,
                bias=False,
            )
        )
        layers.append(nn.ReLU(inplace=True))
        for _ in range(num_of_layers - 2):
            layers.append(
                nn.Conv2d(
                    in_channels=features,
                    out_channels=features,
                    kernel_size=kernel_size,
                    padding=padding,
                    bias=False,
                )
            )
            layers.append(nn.BatchNorm2d(features))
            layers.append(nn.ReLU(inplace=True))
        layers.append(
            nn.Conv2d(
                in_channels=features,
                out_channels=n_channel_out,
                kernel_size=kernel_size,
                padding=padding,
                bias=False,
            )
        )
        self.dncnn = nn.Sequential(*layers)

    def forward(self, x):
        out = self.dncnn(x)
        return out



import inspect
import json
import docstring_parser
import numpy
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QWidget, QGridLayout, QLabel, QCheckBox, QLineEdit

from aydin.util.log.log import lprint


class ConstructorArgumentsWidget(QWidget):
    def __init__(
        self,
        parent,
        arg_names=None,
        arg_defaults=None,
        arg_annotations=None,
        reference_class=None,
    ):
        super(QWidget, self).__init__(parent)
        self.parent = parent

        self.arg_names = arg_names
        self.annotations = []
        self.reference_class = reference_class
        self.line_edits = []

        self.arguments_layout = QGridLayout()
        self.arguments_layout.setAlignment(Qt.AlignTop)

        for index, (name, default_value) in enumerate(zip(arg_names, arg_defaults)):
            param_label = QLabel(f"{name.strip().replace('_', ' ')}: ")
            param_label.setToolTip(
                f"{self.annonation_prettifier(arg_annotations[name])}"
            )

            if default_value is None:
                default_value = "auto"

            if "'bool'" in str(arg_annotations[name]):
                param_linedit = QCheckBox()
                param_linedit.setChecked(default_value)
            elif "dtype" in str(arg_annotations[name]):
                param_linedit = QLineEdit(str(default_value.__name__), self)
            else:
                param_linedit = QLineEdit(str(default_value), self)

            param_linedit.setToolTip(
                f"{self.annonation_prettifier(arg_annotations[name])}"
            )

            if inspect.isclass(reference_class):
                doc = docstring_parser.parse(reference_class.__init__.__doc__)
            else:
                doc = docstring_parser.parse(reference_class.__doc__)
            description = ""
            for param in doc.params:
                if param.arg_name == name:
                    description = param.description

            if description is not None:
                # Handle None to auto replacement:
                description = description.replace("None", "'auto'")
                # Replace new lines with spaces to avoid wrapping conflicts:
                description = description.replace('\n', ' ')

            param_description = QLabel(description)
            param_description.setWordWrap(True)
            param_description.setTextInteractionFlags(Qt.TextSelectableByMouse)

            self.line_edits.append(param_linedit)
            self.annotations.append(arg_annotations[name])

            param_label.setFixedWidth(200)
            param_linedit.setFixedWidth(100)
            self.arguments_layout.addWidget(
                param_label, index, 0, alignment=Qt.AlignTop
            )
            self.arguments_layout.addWidget(
                param_linedit, index, 1, alignment=Qt.AlignTop
            )
            self.arguments_layout.addWidget(
                param_description, index, 2, alignment=Qt.AlignTop
            )

        if not arg_names:
            lprint("No parameters to configure")
            self.arguments_layout.addWidget(QLabel("No parameters to configure"))

        self.setLayout(self.arguments_layout)

    @staticmethod
    def annonation_prettifier(annotation):
        if "'int'" in str(annotation):
            return "int"
        elif "'float'" in str(annotation):
            return "float"
        elif "'bool'" in str(annotation):
            return "bool"
        elif "'str'" in str(annotation):
            return "str"
        else:
            response = str(annotation).replace("typing.", "")
            return response.replace("NoneType", "None")

    @property
    def params_dict(self):
        params_dict = {"class": self.reference_class, "kwargs": {}}

        for name, lineedit, annotation in zip(
            self.arg_names, self.line_edits, self.annotations
        ):

            if lineedit.text() == "auto":
                value = None
            elif "'int'" in str(annotation):
                value = int(lineedit.text())
            elif "'float'" in str(annotation):
                value = float(lineedit.text())
            elif "'bool'" in str(annotation):
                value = bool(lineedit.isChecked())
            elif "'str'" in str(annotation):
                value = lineedit.text()
            elif "dtype" in str(annotation):
                value = numpy.dtype(lineedit.text())
            else:
                value = json.loads(lineedit.text())

            params_dict["kwargs"][name] = value

        return params_dict



import pytest

from aydin.util.log.log import lprint, lsection, Log


@pytest.mark.heavy
def test_log():

    # This is required for this test to pass!
    Log.override_test_exclusion = True

    lprint('Test')

    with lsection('a section'):
        lprint('a line')
        lprint('another line')
        lprint('we are done')

        with lsection('a subsection'):
            lprint('another line')
            lprint('we are done')

            with lsection('a subsection'):
                lprint('another line')
                lprint('we are done')

                assert Log.depth == 3

                with lsection('a subsection'):
                    lprint('another line')
                    lprint('we are done')

                    with lsection('a subsection'):
                        lprint('another line')
                        lprint('we are done')

                        assert Log.depth == 5

                        with lsection('a subsection'):
                            lprint('another line')
                            lprint('we are done')

                            with lsection('a subsection'):
                                lprint('another line')
                                lprint('we are done')

                                assert Log.depth == 7

                        with lsection('a subsection'):
                            lprint('another line')
                            lprint('we are done')

                    with lsection('a subsection'):
                        lprint('another line')
                        lprint('we are done')

                with lsection('a subsection'):
                    lprint('another line')
                    lprint('we are done')

    lprint('test is finished...')

    assert Log.depth == 0



import numpy
import tensorflow as tf
from tensorflow.python.keras import Input
from tensorflow.python.keras.layers import (
    Concatenate,
    Add,
    UpSampling3D,
    Cropping3D,
    ZeroPadding3D,
    UpSampling2D,
    ZeroPadding2D,
    Cropping2D,
    Conv2D,
    Conv3D,
)
from tensorflow.python.keras.models import Model
from tensorflow.python.keras.optimizer_v2.adam import Adam
from tensorflow.python.keras.regularizers import l1

from aydin.nn.layers.util import Split, Rot90
from aydin.nn.layers.maskout import Maskout
from aydin.nn.models.utils.conv_block import (
    conv3d_bn,
    pooling_down3D,
    pooling_down2D,
    conv2d_bn,
)
from aydin.nn.models.utils.training_architectures import get_unet_fit_args
from aydin.nn.util.mask_generator import randmaskgen, maskedgen
from aydin.nn.util.validation_generator import val_data_generator
from aydin.util.log.log import lprint


class UNetModel(Model):
    """UNet model. Three training modes are available: supervised: noisy and clean images are required, shiftconv:
    self-supervised learning with shift and conv scheme non-shiftconv: self-supervised learning by masking pixels at
    each iteration
    """

    def __init__(
        self,
        input_layer_size,
        spacetime_ndim,
        mini_batch_size: int = 1,
        nb_unet_levels: int = 4,
        normalization: str = 'batch',  # None,  # 'instance',
        activation: str = 'ReLU',
        supervised: bool = False,
        shiftconv: bool = True,
        nfilters: int = 8,
        learning_rate: float = 0.01,
        original_zdim: int = None,
        weight_decay: int = 0,
        residual: bool = False,
        pooling_mode: str = 'max',
    ):
        """

        Parameters
        ----------
        input_layer_size
        spacetime_ndim
        mini_batch_size : int
            Mini-batch size
        nb_unet_levels : int
            Depth level of the UNet
        normalization : string
            normalization type, can be `batch` and `instance` for now
        activation : string
            Type of the activation function to use
        supervised : bool
            Flag that controls training approach
        shiftconv : bool
            Flag that controls use of shift convolutions
        nfilters : int
            Number of filters in first layer
        learning_rate : float
            Learning rate
        original_zdim : int
            Original Z-dimension length
        weight_decay : int
            coefficient of l1 regularizer
        residual : bool
            whether to use add or concat at merging layers
        pooling_mode : str
        """
        self.compiled = False

        self.rot_batch_size = mini_batch_size
        self.num_lyr = nb_unet_levels
        self.normalization = normalization
        self.activation = activation
        self.shiftconv = shiftconv
        self.nfilters = nfilters
        self.learning_rate = learning_rate
        self.weight_decay = weight_decay
        self.residual = residual
        self.pooling_mode = pooling_mode

        if type(input_layer_size) is int:
            input_layer_size = (input_layer_size,) * spacetime_ndim + (1,)

        if spacetime_ndim == 3:
            if original_zdim:
                self.zdim = original_zdim
            else:
                self.zdim = input_layer_size[0]

        # Generate a model
        self.input_lyr = Input(input_layer_size, name='input')
        x = (
            self.unet_core_2d(self.input_lyr)
            if spacetime_ndim == 2
            else self.unet_core_3d()
        )

        if not shiftconv and not supervised:
            input_msk = Input(input_layer_size, name='input_msk')
            x = Maskout(name='maskout')([x, input_msk])
            super().__init__([self.input_lyr, input_msk], x)
        else:
            super().__init__(self.input_lyr, x)

        self.compile(optimizer=Adam(lr=learning_rate), loss='mse')
        self.compile = True

    def unet_core_2d(self, input_lyr):
        """Unet Core method which actually populates the model"""
        # Rotation & stack of the input images
        if self.shiftconv:
            input1 = Rot90(input_lyr, kk=1, lyrname='rot1')(input_lyr)
            input2 = Rot90(input_lyr, kk=2, lyrname='rot2')(input_lyr)
            input3 = Rot90(input_lyr, kk=3, lyrname='rot3')(input_lyr)
            x = Concatenate(name='conc_in', axis=0)([input_lyr, input1, input2, input3])
        else:
            x = input_lyr

        skiplyr = [x]
        for i in range(self.num_lyr):
            if i == 0:
                x = conv2d_bn(
                    x,
                    unit=self.nfilters * (i + 1),
                    shiftconv=self.shiftconv,
                    weight_decay=self.weight_decay,
                    lyrname=f'enc{i}_cv0',
                )

            x = conv2d_bn(
                x,
                unit=self.nfilters * (i + 1),
                shiftconv=self.shiftconv,
                norm=self.normalization,
                act=self.activation,
                weight_decay=self.weight_decay,
                lyrname=f'enc{i}',
            )
            x = pooling_down2D(
                x, self.shiftconv, mode=self.pooling_mode, lyrname=f'enc{i}pl'
            )
            if i != self.num_lyr - 1:
                skiplyr.append(x)

        x = conv2d_bn(
            x,
            self.nfilters,
            shiftconv=self.shiftconv,
            norm=self.normalization,
            act=self.activation,
            weight_decay=self.weight_decay,
            lyrname='bottm',
        )

        for i in range(self.num_lyr):
            x = UpSampling2D((2, 2), name=f'up{i}')(x)
            if self.residual:
                x = Add(name=f'add{i}')([x, skiplyr.pop()])
            else:
                x = Concatenate(name=f'cnct{i}')([x, skiplyr.pop()])
            x = conv2d_bn(
                x,
                self.nfilters * max((self.num_lyr - i - 2), 1),
                shiftconv=self.shiftconv,
                weight_decay=self.weight_decay,
                lyrname=f'dec{i}_cv0',
            )
            x = conv2d_bn(
                x,
                self.nfilters * max((self.num_lyr - i - 2), 1),
                shiftconv=self.shiftconv,
                norm=self.normalization,
                act=self.activation,
                weight_decay=self.weight_decay,
                lyrname=f'dec{i}',
            )

        if self.shiftconv:
            # Shift the center pixel
            x = ZeroPadding2D(((0, 0), (1, 0)), name='shiftc_pd')(x)
            x = Cropping2D(((0, 0), (0, 1)), name='shiftc_crp')(x)

            # Rotation & stack for the output
            output0 = Split(x, 0, self.rot_batch_size, 'split0')(x)
            output1 = Split(x, 1, self.rot_batch_size, 'split1')(x)
            output2 = Split(x, 2, self.rot_batch_size, 'split2')(x)
            output3 = Split(x, 3, self.rot_batch_size, 'split3')(x)
            output1 = Rot90(output1, -1, lyrname='rot4')(output1)
            output2 = Rot90(output2, -2, lyrname='rot5')(output2)
            output3 = Rot90(output3, -3, lyrname='rot6')(output3)
            x = Concatenate(name='cnct_last', axis=-1)(
                [output0, output1, output2, output3]
            )
            x = conv2d_bn(
                x,
                self.nfilters * 2 * 4,
                kernel_size=1,
                shiftconv=False,
                norm=self.normalization,
                act=self.activation,
                weight_decay=self.weight_decay,
                lyrname='last1',
            )
            x = conv2d_bn(
                x,
                self.nfilters,
                kernel_size=1,
                shiftconv=False,
                norm=self.normalization,
                act=self.activation,
                weight_decay=self.weight_decay,
                lyrname='last2',
            )

        x = Conv2D(
            1,
            (1, 1),
            padding='same',
            name='last0',
            kernel_regularizer=l1(self.weight_decay),
            bias_regularizer=l1(self.weight_decay),
            activation='linear',
        )(x)

        return x

    def unet_core_3d(self):
        """Unet Core method which actually populates the model"""
        # Rotation & stack of the input images
        if self.shiftconv:
            input1 = Rot90(self.input_lyr, kk=1, lyrname='rot1')(self.input_lyr)
            input2 = Rot90(self.input_lyr, kk=2, lyrname='rot2')(self.input_lyr)
            input3 = Rot90(self.input_lyr, kk=3, lyrname='rot3')(self.input_lyr)
            input5 = Rot90(self.input_lyr, kk=5, lyrname='rot5')(self.input_lyr)
            input6 = Rot90(self.input_lyr, kk=6, lyrname='rot6')(self.input_lyr)
            x = Concatenate(name='conc_in', axis=0)(
                [self.input_lyr, input1, input2, input3, input5, input6]
            )
        else:
            x = self.input_lyr

        skiplyr = [x]
        down2D_n = 0
        for i in range(self.num_lyr):
            if i == 0:
                x = conv3d_bn(
                    x,
                    unit=self.nfilters * (i + 1),
                    shiftconv=self.shiftconv,
                    weight_decay=self.weight_decay,
                    lyrname=f'enc{i}_cv0',
                )

            x = conv3d_bn(
                x,
                unit=self.nfilters * (i + 1),
                shiftconv=self.shiftconv,
                norm=self.normalization,
                act=self.activation,
                weight_decay=self.weight_decay,
                lyrname=f'enc{i}',
            )
            if self.zdim > 3:
                pool_size = (2, 2, 2)
            else:
                if self.shiftconv:
                    raise ValueError(
                        'Input size is too small against the depth of the CNN model. '
                        'Please use masking method or less num_lyr or larger input size.'
                    )
                else:
                    pool_size = (1, 2, 2)
                    down2D_n += 1

            if self.zdim % 2 != 0:
                x = tf.pad(
                    x,
                    tf.constant([[0, 0], [0, 1], [0, 0], [0, 0], [0, 0]]),
                    mode='REFLECT',
                )
                self.zdim = (self.zdim + 1) // 2
            else:
                self.zdim = self.zdim // 2

            x = pooling_down3D(x, self.shiftconv, pool_size, lyrname=f'enc{i}pl')

            if i != self.num_lyr - 1:
                skiplyr.append(x)

        x = conv3d_bn(
            x,
            self.nfilters,
            shiftconv=self.shiftconv,
            norm=self.normalization,
            act=self.activation,
            weight_decay=self.weight_decay,
            lyrname='bottm',  # * num_layer,
        )

        for i in range(self.num_lyr):
            if down2D_n > 0:
                x = UpSampling3D((1, 2, 2), name=f'up{i}')(x)
                down2D_n -= 1
            else:
                x = UpSampling3D((2, 2, 2), name=f'up{i}')(x)
            # Check if z dim of connecting layers from encoder & decoder are the same
            x_en = skiplyr.pop()
            if x_en.shape[1] != x.shape[1]:
                x = Cropping3D(((0, 1), (0, 0), (0, 0)), name=f'dec_crp{i}')(x)
            if self.residual:
                x = Add(name=f'add{i}')([x, x_en])
            else:
                x = Concatenate(name=f'cnct{i}')([x, x_en])
            x = conv3d_bn(
                x,
                self.nfilters * max((self.num_lyr - i - 2), 1),
                shiftconv=self.shiftconv,
                weight_decay=self.weight_decay,
                lyrname=f'dec{i}_cv0',
            )
            x = conv3d_bn(
                x,
                self.nfilters * max((self.num_lyr - i - 2), 1),
                shiftconv=self.shiftconv,
                norm=self.normalization,
                act=self.activation,
                weight_decay=self.weight_decay,
                lyrname=f'dec{i}',
            )

        if self.shiftconv:
            # Shift the center pixel
            x = ZeroPadding3D(((0, 0), (0, 0), (1, 0)), name='shiftc_pd')(x)
            x = Cropping3D(((0, 0), (0, 0), (0, 1)), name='shiftc_crp')(x)

            # Rotation & stack for the output
            output0 = Split(x, 0, self.rot_batch_size, 'split0')(x)
            output1 = Split(x, 1, self.rot_batch_size, 'split1')(x)
            output2 = Split(x, 2, self.rot_batch_size, 'split2')(x)
            output3 = Split(x, 3, self.rot_batch_size, 'split3')(x)
            output5 = Split(x, 4, self.rot_batch_size, 'split5')(x)
            output6 = Split(x, 5, self.rot_batch_size, 'split6')(x)
            output1 = Rot90(output1, -1, lyrname='rot-1')(output1)
            output2 = Rot90(output2, -2, lyrname='rot-2')(output2)
            output3 = Rot90(output3, -3, lyrname='rot-3')(output3)
            output5 = Rot90(output5, -5, lyrname='rot-5')(output5)
            output6 = Rot90(output6, -6, lyrname='rot-6')(output6)
            x = Concatenate(name='cnct_last', axis=-1)(
                [output0, output1, output2, output3, output5, output6]
            )
            x = conv3d_bn(
                x,
                self.nfilters * 2 * 4,
                kernel_size=3,  # a work around for a bug in tf; supposed to be 1
                shiftconv=False,
                norm=self.normalization,
                act=self.activation,
                weight_decay=self.weight_decay,
                lyrname='last1',
            )
            x = conv3d_bn(
                x,
                self.nfilters,
                kernel_size=3,  # a work around for a bug in tf; supposed to be 1
                shiftconv=False,
                norm=self.normalization,
                act=self.activation,
                weight_decay=self.weight_decay,
                lyrname='last2',
            )

        x = Conv3D(
            1,
            3,  # a work around for a bug in tf; supposed to be 1
            padding='same',
            name='last0',
            kernel_regularizer=l1(self.weight_decay),
            bias_regularizer=l1(self.weight_decay),
            activation='linear',
        )(x)

        return x

    def size(self):
        """Returns size of the model in bytes"""
        return self.count_params() * 4

    def fit(
        self,
        input_image=None,
        target_image=None,
        callbacks=None,
        train_valid_ratio=None,
        img_val=None,
        val_marker=None,
        training_architecture=None,
        create_patches_for_validation=None,
        total_num_patches=None,
        batch_size=None,
        random_mask_ratio=None,
        patch_size=None,
        mask_size=None,
        replace_by="zero",
        verbose=None,
        max_epochs=None,
        ReduceLR_patience=None,
        reduce_lr_factor=0.3,
        parent=None,
    ):
        """

        Parameters
        ----------
        input_image
        target_image
        callbacks
        train_valid_ratio
        img_val
        val_marker
        training_architecture
        create_patches_for_validation
        total_num_patches
        batch_size
        random_mask_ratio
        patch_size
        mask_size
        replace_by
        verbose
        max_epochs
        ReduceLR_patience
        reduce_lr_factor
        parent

        Returns
        -------
        loss_history

        """
        if create_patches_for_validation:
            tv_ratio = train_valid_ratio
        else:
            tv_ratio = 0

        validation_data, validation_steps = get_unet_fit_args(
            train_method=training_architecture,
            create_patches_for_validation=create_patches_for_validation,
            input_image=input_image,
            total_num_patches=total_num_patches,
            train_valid_ratio=train_valid_ratio,
            batch_size=batch_size,
            random_mask_ratio=random_mask_ratio,
            img_val=img_val,
            patch_size=patch_size,
            mask_size=mask_size,
            val_marker=val_marker,
            replace_by=replace_by,
        )

        if 'shiftconv' in training_architecture:
            loss_history = super().fit(
                input_image,
                target_image,
                epochs=max_epochs,
                callbacks=callbacks,
                verbose=verbose,
                batch_size=batch_size,
                validation_data=validation_data,
            )
        elif 'checkerbox' in training_architecture:
            loss_history = super().fit(
                maskedgen(input_image, batch_size, mask_size, replace_by=replace_by),
                epochs=max_epochs,
                steps_per_epoch=numpy.prod(mask_size)
                * numpy.ceil(input_image.shape[0] * (1 - tv_ratio) / batch_size).astype(
                    int
                ),
                verbose=verbose,
                callbacks=callbacks,
                validation_data=validation_data,
                validation_steps=validation_steps,
            )
        elif 'random' in training_architecture:
            loss_history = super().fit(
                randmaskgen(
                    input_image,
                    batch_size,
                    p_maskedpixels=random_mask_ratio,
                    replace_by=replace_by,
                ),
                epochs=max_epochs,
                steps_per_epoch=numpy.ceil(1 / random_mask_ratio).astype(int)
                * numpy.ceil(total_num_patches * (1 - tv_ratio) / batch_size).astype(
                    int
                ),
                verbose=verbose,
                callbacks=callbacks,
                validation_data=validation_data,
                validation_steps=validation_steps,
            )
        elif 'checkran' in training_architecture:
            # train with checkerbox first
            lprint('Starting with checkerbox masking.')
            history_checkerbox = super().fit(
                maskedgen(input_image, batch_size, mask_size, replace_by=replace_by),
                epochs=max_epochs,
                steps_per_epoch=numpy.prod(mask_size)
                * numpy.ceil(input_image.shape[0] * (1 - tv_ratio) / batch_size).astype(
                    int
                ),
                verbose=verbose,
                callbacks=callbacks[:-2] + callbacks[-1:],
                validation_data=maskedgen(
                    input_image, batch_size, mask_size, replace_by=replace_by
                )
                if create_patches_for_validation
                else val_data_generator(
                    input_image,
                    img_val,
                    val_marker,
                    batch_size,
                    train_valid_ratio=train_valid_ratio,
                ),
                validation_steps=max(
                    numpy.ceil(
                        input_image.shape[0] * train_valid_ratio / batch_size
                    ).astype(int),
                    1,
                ),
            )

            # Then switch to random masking
            lprint('Switched to random masking.')
            history_random = super().fit(
                randmaskgen(
                    input_image,
                    batch_size,
                    p_maskedpixels=random_mask_ratio,
                    replace_by=replace_by,
                ),
                epochs=max_epochs,
                steps_per_epoch=numpy.ceil(1 / random_mask_ratio).astype(int)
                * numpy.ceil(total_num_patches * (1 - tv_ratio) / batch_size).astype(
                    int
                ),
                verbose=verbose,
                callbacks=callbacks[:-1],
                initial_epoch=history_checkerbox.epoch[-1] + 1,
                validation_data=randmaskgen(
                    input_image,
                    batch_size,
                    p_maskedpixels=random_mask_ratio,
                    replace_by=replace_by,
                )
                if create_patches_for_validation
                else val_data_generator(
                    input_image,
                    img_val,
                    val_marker,
                    batch_size,
                    train_valid_ratio=train_valid_ratio,
                ),
                validation_steps=max(
                    numpy.ceil(
                        total_num_patches * train_valid_ratio / batch_size
                    ).astype(int),
                    1,
                ),
            )
            history_checkerbox.epoch += history_random.epoch
            for key, val in history_random.history.items():
                history_checkerbox.history[key] += history_random.history[key]
            loss_history = history_checkerbox
        else:
            loss_history = -1

        return loss_history



import numpy

from aydin.analysis.camera_simulation import simulate_camera_image
from aydin.io.datasets import characters
from aydin.it.transforms.variance_stabilisation import VarianceStabilisationTransform


def demo_vst():

    image = characters()
    image = image.astype(numpy.float32) * 0.1
    noisy = simulate_camera_image(image)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(image, name='noisy')

        for mode in [
            'yeo-johnson',
            'box-cox',
            'quantile',
            'anscomb',
            'log',
            'sqrt',
            'identity',
        ]:
            print(f"testing mode: {mode}")
            vst = VarianceStabilisationTransform(mode=mode)

            preprocessed = vst.preprocess(noisy)
            postprocessed = vst.postprocess(preprocessed)

            viewer.add_image(preprocessed, name='preprocessed_' + mode)
            viewer.add_image(postprocessed, name='postprocessed_' + mode)


demo_vst()



import numpy
from numpy.linalg import norm
from scipy.fft import idstn

from aydin.features.groups.convolutional import ConvolutionalFeatures


class DCTFeatures(ConvolutionalFeatures):
    """
    DCT Feature Group class
    """

    def __init__(self, size: int, max_freq: float = 0.75, power: float = 0.5):
        super().__init__(kernels=None)
        self.size = size
        self.max_freq = max_freq
        self.power = power

        self.image = None
        self.exclude_center: bool = False

    def _ensure_dct_kernels_available(self, ndim: int):
        # Ensures that the kernels are available for subsequent steps.
        # We can't construct the kernels until we know the dimension of the image
        if self.kernels is None or self.kernels[0].ndim != ndim:
            dct_kernels = []
            shape = tuple((self.size,) * ndim)
            dctcoefs = numpy.zeros(shape=shape)

            for index, _ in numpy.ndenumerate(dctcoefs):

                freqs = numpy.array([i / self.size for i in index])
                freq = norm(freqs)

                if freq > self.max_freq:
                    continue

                dctcoefs[...] = 0
                dctcoefs[index] = 1

                kernel = idstn(dctcoefs, norm="ortho")
                kernel /= kernel.std()
                kernel = numpy.sign(kernel) * abs(kernel) ** self.power
                kernel /= kernel.std()
                kernel += kernel.min()
                kernel /= kernel.sum()
                kernel = kernel.astype(numpy.float32)

                dct_kernels.append(kernel)

            self.kernels = dct_kernels

    @property
    def receptive_field_radius(self) -> int:
        return self.size // 2

    def num_features(self, ndim: int) -> int:
        self._ensure_dct_kernels_available(ndim)
        return super().num_features(ndim)

    def prepare(self, image, excluded_voxels=None, **kwargs):
        if excluded_voxels is None:
            excluded_voxels = []

        self._ensure_dct_kernels_available(image.ndim)
        super().prepare(image, excluded_voxels, **kwargs)



from qtpy.QtWidgets import QWidget, QHBoxLayout

from aydin.gui._qt.output_wrapper import OutputWrapper
from aydin.gui._qt.job_runners.worker import Worker
from aydin.io.io import imwrite
from aydin.io.utils import (
    get_output_image_path,
    get_options_json_path,
    get_save_model_path,
)
from aydin.restoration.denoise.util.denoise_utils import get_denoiser_class_instance
from aydin.util.log.log import Log, lprint


class DenoiseJobRunner(QWidget):
    def __init__(self, parent, threadpool, start_button):
        super(QWidget, self).__init__(parent)
        self.parent = parent
        self.threadpool = threadpool
        self.start_button = start_button

        stdout = OutputWrapper(self, True)
        stdout.outputWritten.connect(self.progress_fn)
        stderr = OutputWrapper(self, False)
        stderr.outputWritten.connect(self.progress_fn)

        self.result_images = []
        self.early_stopped = False

        self.widget_layout = QHBoxLayout()

        self.start_button.clicked.connect(self.prep_and_run)
        # self.widget_layout.addWidget(self.start_button)

        # self.stop_button = QPushButton(
        #     "Stop", icon=QApplication.style().standardIcon(QStyle.SP_MediaStop)
        # )
        # self.stop_button.clicked.connect(self.stop_func)
        # self.widget_layout.addWidget(self.stop_button)

        self.setLayout(self.widget_layout)

    def stop_running(self):
        self.early_stopped = True
        self.denoiser.stop_running()

    def start_func(self, progress_callback):
        Log.gui_callback = progress_callback

        results = []

        for training_image, inference_image, image_path in zip(
            self.training_images, self.inference_images, self.image_paths
        ):
            self.denoiser.train(
                training_image, batch_axes=self.batch_axes, chan_axes=self.channel_axes
            )

            if self.denoiser.it:
                denoised = self.denoiser.denoise(
                    inference_image,
                    batch_axes=self.batch_axes,
                    chan_axes=self.channel_axes,
                )
                results.append(denoised)

                output_path, file_counter = get_output_image_path(image_path)

                imwrite(denoised, output_path)
                if self.save_options_json:
                    json_path = get_options_json_path(
                        image_path, passed_counter=file_counter
                    )
                    self.parent.save_options_json(json_path)
                    lprint(f"DONE, options json written in {json_path}")

                if self.save_model:
                    model_path = get_save_model_path(
                        image_path, passed_counter=file_counter
                    )
                    self.denoiser.save_model(model_path)
                    lprint(f"DONE, trained model written in {model_path}")

                lprint(f"DONE, results written in {output_path}")
            else:
                self.early_stopped = True
                lprint("DONE, failed to run...")

        Log.gui_callback = None
        return results

    def progress_fn(self, log_str):
        self.parent.activity_widget.infoTextBox.insertPlainText(log_str)

    def result_callback(self, results):
        self.result_images += results

    def thread_complete(self):
        self.start_button.setEnabled(True)

        # Turn the overlay off
        self.parent.overlay.hide()

        # Open napari with input and output images
        if not self.early_stopped:
            self.parent.open_images_with_napari()

    def prep_and_run(self):
        # Get images and their related data
        self.image_paths = [i[5] for i in self.parent.data_model.images_to_denoise]
        if len(self.image_paths) == 0:
            lprint("Aydin cannot be started with no image")
            return

        self.training_images = self.parent.tabs["Training Crop"].images
        self.inference_images = (
            self.parent.tabs["Denoising Crop"].images
            if self.parent.tabwidget.isTabEnabled(
                self.parent.tabwidget.indexOf(self.parent.tabs["Denoising Crop"])
            )
            else self.parent.tabs["Training Crop"].images
        )
        self.batch_axes = self.parent.tabs["Dimensions"].batch_axes
        self.channel_axes = self.parent.tabs["Dimensions"].channel_axes
        self.denoise_backend = self.parent.tabs["Denoise"].selected_backend
        self.it_transforms = self.parent.tabs["Pre/Post-Processing"].transforms

        self.lower_level_args = self.parent.tabs["Denoise"].lower_level_args

        self.save_options_json = self.parent.tabs[
            "Denoise"
        ].current_backend_qwidget.save_json_checkbox.isChecked()

        self.save_model = self.parent.tabs[
            "Denoise"
        ].current_backend_qwidget.save_model_checkbox.isChecked()

        # Initialize required restoration instances
        self.denoiser = get_denoiser_class_instance(
            variant=self.denoise_backend,
            lower_level_args=self.lower_level_args,
            it_transforms=self.it_transforms,
        )

        Log.gui_statusbar = self.parent.parent.statusBar

        # Show activity widget
        self.parent.activity_dock.setHidden(False)

        # Turn the overlay on
        self.parent.overlay.show()

        # Pass the function to execute
        worker = Worker(
            self.start_func
        )  # Any other args, kwargs are passed to the run function

        worker.signals.result.connect(self.result_callback)
        worker.signals.finished.connect(self.thread_complete)
        worker.signals.progress.connect(self.progress_fn)

        self.parent.activity_widget.clear_activity()

        self.result_images = []

        self.start_button.setEnabled(False)

        # Execute
        self.threadpool.start(worker)

    def result_image_arrays(self):
        return self.result_images



import cmath
import math
import numpy
import scipy
import scipy.integrate
import scipy.interpolate
import scipy.special


class MicroscopePSF:
    """
    Generate a PSF using the Gibson and Lanni model.

    Note: All distance units are microns.

    This is slightly reworked version of the Python code provided by Kyle
    Douglass, "Implementing a fast Gibson-Lanni PSF solver in Python".

    http://kmdouglass.github.io/posts/implementing-a-fast-gibson-lanni-psf-solver-in-python.html

    Object refactored by Loic Royer,
    Original here: https://github.com/MicroscPSF/MicroscPSF-Py

    References:

    1. Li et al, "Fast and accurate three-dimensional point spread function computation
       for fluorescence microscopy", JOSA, 2017.

    2. Gibson, S. & Lanni, F. "Experimental test of an analytical model of
       aberration in an oil-immersion objective lens used in three-dimensional
       light microscopy", J. Opt. Soc. Am. A 9, 154-166 (1992), [Originally
       published in J. Opt. Soc. Am. A 8, 1601-1613 (1991)].

    3. Kirshner et al, "3-D PSF fitting for fluorescence microscopy: implementation
       and localization application", Journal of Microscopy, 2012.

    Hazen 04/18
    """

    def __init__(self):
        """ """

        # Internal constants.
        self.num_basis = (
            100  # Number of rescaled Bessels that approximate the phase function.
        )
        self.rho_samples = 1000  # Number of pupil sample along the radial direction.

        # Microscope parameters.
        # IMPORTANT: DO NOT CHANGE THESE DEFAULTS, IMPORTANT FOR TESTS
        self.parameters = {
            "M": 100.0,  # magnification
            "NA": 1.4,  # numerical aperture
            "ng0": 1.515,  # coverslip RI design value
            "ng": 1.515,  # coverslip RI experimental value
            "ni0": 1.515,  # immersion medium RI design value
            "ni": 1.515,  # immersion medium RI experimental value
            "ns": 1.33,  # specimen refractive index (RI)
            "ti0": 150,  # microns, working distance (immersion medium thickness) design value
            "tg": 170,  # microns, coverslip thickness experimental value
            "tg0": 170,  # microns, coverslip thickness design value
            "zd0": 200.0 * 1.0e3,
        }  # microscope tube length (in microns).

    def _calcRv(self, dxy, xy_size, sampling=2):
        """
        Calculate rv vector, this is 2x up-sampled.
        """
        rv_max = math.sqrt(0.5 * xy_size * xy_size) + 1
        return numpy.arange(0, rv_max * dxy, dxy / sampling)

    def _configure(self, wvl):

        mp = self.parameters

        # Scaling factors for the Fourier-Bessel series expansion
        min_wavelength = 0.436  # microns
        scaling_factor = (
            mp["NA"]
            * (3 * numpy.arange(1, self.num_basis + 1) - 2)
            * min_wavelength
            / wvl
        )

        # Not sure this is completely correct for the case where the axial
        # location of the flourophore is 0.0.
        #
        max_rho = (
            min([mp["NA"], mp["ng0"], mp["ng"], mp["ni0"], mp["ni"], mp["ns"]])
            / mp["NA"]
        )

        return [scaling_factor, max_rho]

    def deltaFocus(self, zd):
        """
        Return focal offset needed to compensate for the camera being at zd.

        mp - The microscope parameters dictionary.
        zd - Actual camera position in microns.
        """
        mp = self.parameters

        a = mp["NA"] * mp["zd0"] / mp["M"]  # Aperture radius at the back focal plane.
        return a * a * (mp["zd0"] - zd) / (2.0 * mp["zd0"] * zd)

    def gLXYZCameraScan(
        self, dxy, xy_size, zd, normalize=True, pz=0.0, wvl=0.6, zv=0.0
    ):
        """
        NOTE: Does not work!

        Calculate 3D G-L PSF. This is models the PSF you would measure by scanning the
        camera position (changing the microscope tube length).

        This will return a numpy array with of size (zv.size, xy_size, xy_size). Note that z
        is the zeroth dimension of the PSF.

        dxy - Step size in the XY plane.
        xy_size - Number of pixels in X/Y.
        zd - A numpy array containing the camera positions in microns.

        normalize - Normalize the PSF to unit height.
        pz - Particle z position above the coverslip (positive values only).
        wvl - Light wavelength in microns.
        zv - The (relative) z offset value of the coverslip (negative is closer to the objective).
        """
        # Calculate rv vector, this is 2x up-sampled.
        rv = self._calcRv(dxy, xy_size)

        # Calculate radial/Z PSF.
        PSF_rz = self.gLZRCameraScan(rv, zd, normalize=normalize, pz=pz, wvl=wvl, zv=zv)

        # Create XYZ PSF by interpolation.
        return self.psfRZToPSFXYZ(dxy, xy_size, rv, PSF_rz)

    def gLXYZFocalScan(
        self, dxy, xy_size, zv, normalize=True, pz=0.0, wvl=0.6, zd=None
    ):
        """
        Calculate 3D G-L PSF. This is models the PSF you would measure by scanning the microscopes
        focus.

        This will return a numpy array with of size (zv.size, xy_size, xy_size). Note that z
        is the zeroth dimension of the PSF.

        dxy - Step size in the XY plane.
        xy_size - Number of pixels in X/Y.
        zv - A numpy array containing the (relative) z offset values of the coverslip (negative is closer to the objective).

        normalize - Normalize the PSF to unit height.
        pz - Particle z position above the coverslip (positive values only).
        wvl - Light wavelength in microns.
        zd - Actual camera position in microns. If not specified the microscope tube length is used.
        """
        # Calculate rv vector, this is 2x up-sampled.
        rv = self._calcRv(dxy, xy_size)

        # Calculate radial/Z PSF.
        PSF_rz = self.gLZRFocalScan(rv, zv, normalize=normalize, pz=pz, wvl=wvl, zd=zd)

        # Create XYZ PSF by interpolation.
        return self.psfRZToPSFXYZ(dxy, xy_size, rv, PSF_rz)

    def gLXYZParticleScan(
        self, dxy, xy_size, pz, normalize=True, wvl=0.6, zd=None, zv=0.0
    ):
        """
        Calculate 3D G-L PSF. This is models the PSF you would measure by scanning a particle
        through the microscopes focus.

        This will return a numpy array with of size (zv.size, xy_size, xy_size). Note that z
        is the zeroth dimension of the PSF.

        dxy - Step size in the XY plane.
        xy_size - Number of pixels in X/Y.
        pz - A numpy array containing the particle z position above the coverslip (positive values only)
             in microns.

        normalize - Normalize the PSF to unit height.
        wvl - Light wavelength in microns.
        zd - Actual camera position in microns. If not specified the microscope tube length is used.
        zv - The (relative) z offset value of the coverslip (negative is closer to the objective).
        """
        # Calculate rv vector, this is 2x up-sampled.
        rv = self._calcRv(dxy, xy_size)

        # Calculate radial/Z PSF.
        PSF_rz = self.gLZRParticleScan(
            rv, pz, normalize=normalize, wvl=wvl, zd=zd, zv=zv
        )

        # Create XYZ PSF by interpolation.
        return self.psfRZToPSFXYZ(dxy, xy_size, rv, PSF_rz)

    def gLZRScan(self, pz, rv, zd, zv, normalize=True, wvl=0.6):
        """
        Calculate radial G-L at specified radius. This function is primarily designed
        for internal use. Note that only one pz, zd and zv should be a numpy array
        with more than one element. You can simulate scanning the focus, the particle
        or the camera but not 2 or 3 of these values at the same time.

        mp - The microscope parameters dictionary.
        pz - A numpy array containing the particle z position above the coverslip (positive values only).
        rv - A numpy array containing the radius values.
        zd - A numpy array containing the actual camera position in microns.
        zv - A numpy array containing the relative z offset value of the coverslip (negative is closer to the objective).

        normalize - Normalize the PSF to unit height.
        wvl - Light wavelength in microns.
        """

        mp = self.parameters

        [scaling_factor, max_rho] = self._configure(wvl)
        rho = numpy.linspace(0.0, max_rho, self.rho_samples)

        a = (
            mp["NA"] * mp["zd0"] / math.sqrt(mp["M"] * mp["M"] + mp["NA"] * mp["NA"])
        )  # Aperture radius at the back focal plane.
        k = 2.0 * numpy.pi / wvl

        ti = zv.reshape(-1, 1) + mp["ti0"]
        pz = pz.reshape(-1, 1)
        zd = zd.reshape(-1, 1)

        opdt = self.OPD(rho, ti, pz, wvl, zd)

        # Sample the phase
        # phase = numpy.cos(opdt) + 1j * numpy.sin(opdt)
        phase = numpy.exp(1j * opdt)

        # Define the basis of Bessel functions
        # Shape is (number of basis functions by number of rho samples)
        J = scipy.special.jv(0, scaling_factor.reshape(-1, 1) * rho)

        # Compute the approximation to the sampled pupil phase by finding the least squares
        # solution to the complex coefficients of the Fourier-Bessel expansion.
        # Shape of C is (number of basis functions by number of z samples).
        # Note the matrix transposes to get the dimensions correct.
        C, residuals, _, _ = numpy.linalg.lstsq(J.T, phase.T, rcond=None)

        rv = rv * mp["M"]
        b = k * a * rv.reshape(-1, 1) / zd

        # Convenience functions for J0 and J1 Bessel functions
        J0 = lambda x: scipy.special.jv(0, x)  # noqa: E731
        J1 = lambda x: scipy.special.jv(1, x)  # noqa: E731

        # See equation 5 in Li, Xue, and Blu
        denom = scaling_factor * scaling_factor - b * b
        R = (
            scaling_factor * J1(scaling_factor * max_rho) * J0(b * max_rho) * max_rho
            - b * J0(scaling_factor * max_rho) * J1(b * max_rho) * max_rho
        )
        R /= denom

        # The transpose places the axial direction along the first dimension of the array, i.e. rows
        # This is only for convenience.
        PSF_rz = (numpy.abs(R.dot(C)) ** 2).T

        # Normalize to the maximum value
        if normalize:
            PSF_rz /= numpy.max(PSF_rz)

        return PSF_rz

    def gLZRCameraScan(self, rv, zd, normalize=True, pz=0.0, wvl=0.6, zv=0.0):
        """
        NOTE: Does not work!

        Calculate radial G-L at specified radius and z values. This is models the PSF
        you would measure by scanning the camera position (changing the microscope
        tube length).

        rv - A numpy array containing the radius values.
        zd - A numpy array containing the camera positions in microns.

        normalize - Normalize the PSF to unit height.
        pz - Particle z position above the coverslip (positive values only).
        wvl - Light wavelength in microns.
        zv - The (relative) z offset value of the coverslip (negative is closer to the objective).
        """
        pz = numpy.array([pz])
        zv = numpy.array([zv])

        return self.gLZRScan(pz, rv, zd, zv, normalize=normalize, wvl=wvl)

    def gLZRFocalScan(self, rv, zv, normalize=True, pz=0.0, wvl=0.6, zd=None):
        """
        Calculate radial G-L at specified radius and z values. This is models the PSF
        you would measure by scanning the microscopes focus.

        mp - The microscope parameters dictionary.
        rv - A numpy array containing the radius values.
        zv - A numpy array containing the (relative) z offset values of the coverslip (negative is
             closer to the objective) in microns.

        normalize - Normalize the PSF to unit height.
        pz - Particle z position above the coverslip (positive values only).
        wvl - Light wavelength in microns.
        zd - Actual camera position in microns. If not specified the microscope tube length is used.
        """

        mp = self.parameters

        if zd is None:
            zd = mp["zd0"]

        pz = numpy.array([pz])
        zd = numpy.array([zd])

        return self.gLZRScan(pz, rv, zd, zv, normalize=normalize, wvl=wvl)

    def gLZRParticleScan(self, rv, pz, normalize=True, wvl=0.6, zd=None, zv=0.0):
        """
        Calculate radial G-L at specified radius and z values. This is models the PSF
        you would measure by scanning the particle relative to the microscopes focus.

        mp - The microscope parameters dictionary.
        rv - A numpy array containing the radius values.
        pz - A numpy array containing the particle z position above the coverslip (positive values only)
             in microns.

        normalize - Normalize the PSF to unit height.
        wvl - Light wavelength in microns.
        zd - Actual camera position in microns. If not specified the microscope tube length is used.
        zv - The (relative) z offset value of the coverslip (negative is closer to the objective).
        """

        mp = self.parameters

        if zd is None:
            zd = mp["zd0"]

        zd = numpy.array([zd])
        zv = numpy.array([zv])

        return self.gLZRScan(pz, rv, zd, zv, normalize=normalize, wvl=wvl)

    def OPD(self, rho, ti, pz, wvl, zd):
        """
        Calculate phase aberration term.

        mp - The microscope parameters dictionary.
        rho - Rho term.
        ti - Coverslip z offset in microns.
        pz - Particle z position above the coverslip in microns.
        wvl - Light wavelength in microns.
        zd - Actual camera position in microns.
        """

        mp = self.parameters

        NA = mp["NA"]
        ns = mp["ns"]
        ng0 = mp["ng0"]
        ng = mp["ng"]
        ni0 = mp["ni0"]
        ni = mp["ni"]
        ti0 = mp["ti0"]
        tg = mp["tg"]
        tg0 = mp["tg0"]
        zd0 = mp["zd0"]

        a = NA * zd0 / mp["M"]  # Aperture radius at the back focal plane.
        k = 2.0 * numpy.pi / wvl  # Wave number of emitted light.

        OPDs = pz * numpy.sqrt(ns * ns - NA * NA * rho * rho)  # OPD in the sample.
        OPDi = ti * numpy.sqrt(ni * ni - NA * NA * rho * rho) - ti0 * numpy.sqrt(
            ni0 * ni0 - NA * NA * rho * rho
        )  # OPD in the immersion medium.
        OPDg = tg * numpy.sqrt(ng * ng - NA * NA * rho * rho) - tg0 * numpy.sqrt(
            ng0 * ng0 - NA * NA * rho * rho
        )  # OPD in the coverslip.
        OPDt = (
            a * a * (zd0 - zd) * rho * rho / (2.0 * zd0 * zd)
        )  # OPD in camera position.

        return k * (OPDs + OPDi + OPDg + OPDt)

    def psfRZToPSFXYZ(self, dxy, xy_size, rv, PSF_rz):
        """
        Use interpolation to create a 3D XYZ PSF from a 2D ZR PSF.
        """
        # Create XY grid of radius values.
        c_xy = float(xy_size) * 0.5
        xy = numpy.mgrid[0:xy_size, 0:xy_size] + 0.5
        r_pixel = dxy * numpy.sqrt(
            (xy[1] - c_xy) * (xy[1] - c_xy) + (xy[0] - c_xy) * (xy[0] - c_xy)
        )

        # Create XYZ PSF by interpolation.
        PSF_xyz = numpy.zeros((PSF_rz.shape[0], xy_size, xy_size))
        for i in range(PSF_rz.shape[0]):
            psf_rz_interp = scipy.interpolate.interp1d(rv, PSF_rz[i, :])
            PSF_xyz[i, :, :] = psf_rz_interp(r_pixel.ravel()).reshape(xy_size, xy_size)

        return PSF_xyz

    def slowGL(self, max_rho, rv, zv, pz, wvl, zd):
        """
        Calculate a single point in the G-L PSF using integration. This
        is primarily provided for testing / reference purposes. As the
        function name implies, this is going to be slow.

        mp - The microscope parameters dictionary.
        max_rho - The maximum rho value.
        rv - A radius value in microns.
        zv - A z offset value (of the coverslip) in microns.
        pz - Particle z position above the coverslip in microns.
        wvl - Light wavelength in microns.
        zd - Actual camera position in microns.
        """

        mp = self.parameters

        a = (
            mp["NA"] * mp["zd0"] / math.sqrt(mp["M"] * mp["M"] + mp["NA"] * mp["NA"])
        )  # Aperture radius at the back focal plane.
        k = 2.0 * numpy.pi / wvl
        ti = zv + mp["ti0"]

        rv = rv * mp["M"]

        def integral_fn_imag(rho):
            t1 = k * a * rho * rv / zd
            t2 = scipy.special.jv(0, t1)
            t3 = t2 * cmath.exp(1j * self.OPD(rho, ti, pz, wvl, zd)) * rho
            return t3.imag

        def integral_fn_real(rho):
            t1 = k * a * rho * rv / zd
            t2 = scipy.special.jv(0, t1)
            t3 = t2 * cmath.exp(1j * self.OPD(rho, ti, pz, wvl, zd)) * rho
            return t3.real

        int_i = scipy.integrate.quad(lambda x: integral_fn_imag(x), 0.0, max_rho)[0]
        int_r = scipy.integrate.quad(lambda x: integral_fn_real(x), 0.0, max_rho)[0]

        t1 = k * a * a / (zd * zd)
        return t1 * (int_r * int_r + int_i * int_i)

    def gLZRFocalScanSlow(self, rv, zv, normalize=True, pz=0.0, wvl=0.6, zd=None):
        """
        This is the integration version of gLZRFocalScan.

        mp - The microscope parameters dictionary.
        rv - A numpy array containing the radius values.
        zv - A numpy array containing the (relative) z offset values of the coverslip (negative is closer to the objective).

        normalize - Normalize the PSF to unit height.
        pz - Particle z position above the coverslip (positive values only).
        wvl - Light wavelength in microns.
        zd - Actual camera position in microns. If not specified the microscope tube length is used.
        """

        mp = self.parameters

        if zd is None:
            zd = mp["zd0"]

        [scaling_factor, max_rho] = self._configure(wvl)

        psf_rz = numpy.zeros((zv.size, rv.size))
        for i in range(zv.size):
            for j in range(rv.size):
                psf_rz[i, j] = self.slowGL(max_rho, rv[j], zv[i], pz, wvl, zd)

        if normalize:
            psf_rz = psf_rz / numpy.max(psf_rz)

        return psf_rz

    def gLZRParticleScanSlow(self, rv, pz, normalize=True, wvl=0.6, zd=None, zv=0.0):
        """
        This is the integration version of gLZRParticleScan.

        mp - The microscope parameters dictionary.
        rv - A numpy array containing the radius values.
        pz - A numpy array containing the particle z position above the coverslip (positive values only)
             in microns.

        normalize - Normalize the PSF to unit height.
        wvl - Light wavelength in microns.
        zd - Actual camera position in microns. If not specified the microscope tube length is used.
        zv - The (relative) z offset value of the coverslip (negative is closer to the objective).
        """

        mp = self.parameters

        if zd is None:
            zd = mp["zd0"]

        [scaling_factor, max_rho] = self._configure(wvl)

        psf_rz = numpy.zeros((pz.size, rv.size))
        for i in range(pz.size):
            for j in range(rv.size):
                psf_rz[i, j] = self.slowGL(max_rho, rv[j], zv, pz[i], wvl, zd)

        if normalize:
            psf_rz = psf_rz / numpy.max(psf_rz)

        return psf_rz



import numpy

from aydin.io.datasets import camera, normalise
from aydin.it.balancing.data_histogram_balancer import DataHistogramBalancer


def test_no_balancing():
    balancer = DataHistogramBalancer(keep_ratio=0.5, balance=False)

    image = normalise(camera().astype(numpy.float32, copy=False)).ravel()

    balancer.calibrate(image, batch_length=16)

    entries = [image[i : i + 16] for i in range(0, image.size, 16)]

    balancer.initialise(len(entries))

    count_accepted = 0

    for entry in entries:

        accepted = balancer.add_entry(entry)

        if accepted:
            count_accepted += 1

    print(f"accepted: {count_accepted} / {len(entries)}")
    assert (0.5 - count_accepted / len(entries)) < 0.01


def test_balancing():
    balancer = DataHistogramBalancer(keep_ratio=0.5, balance=True)

    image = normalise(camera().astype(numpy.float32)).ravel()

    balancer.calibrate(image, batch_length=16)

    entries = [image[i : i + 16] for i in range(0, image.size, 16)]

    balancer.initialise(len(entries))

    count_accepted = 0

    for entry in entries:

        accepted = balancer.add_entry(entry)

        if accepted:
            count_accepted += 1

    print(f"accepted: {count_accepted} / {len(entries)}")
    assert count_accepted / len(entries) < 0.4


def test_multiple_runs():
    balancer = DataHistogramBalancer(keep_ratio=0.5, balance=True)

    image = normalise(camera().astype(numpy.float32)).ravel()

    balancer.calibrate(image, batch_length=16)

    entries = [image[i : i + 16] for i in range(0, image.size, 16)]

    for j in range(10):
        balancer.initialise(len(entries))

        count_accepted = 0

        for entry in entries:

            accepted = balancer.add_entry(entry)

            if accepted:
                count_accepted += 1

        print(f"accepted: {count_accepted} / {len(entries)}")
        assert count_accepted / len(entries) < 0.4



from aydin.io.datasets import cropped_newyork

from aydin.it.classic_denoisers.demo.demo_2D_gm import demo_gm
from aydin.it.classic_denoisers.gm import denoise_gm
from aydin.it.classic_denoisers.test.util_test_nd import check_nd


def test_gm():
    assert demo_gm(cropped_newyork(), display=False) >= 0.610 - 0.03


def test_gm_nd():
    check_nd(denoise_gm)



import itertools
import math
import traceback
from copy import copy
from typing import Callable, List, Tuple, Union
import numpy
from joblib import Parallel, delayed
from numpy.linalg import LinAlgError, norm
from scipy.interpolate import RBFInterpolator
from scipy.optimize import minimize

from aydin.util.log.log import lprint, lsection


class Optimizer:
    def __init__(self):
        pass

    def optimize(
        self,
        function: Callable,
        bounds: List[Union[Tuple[int, ...], Tuple[float, ...]]],
        init_strategies: str = 'corners+centers+random',
        exploration_rate: float = 0.4,
        patience: int = 64,
        max_num_evaluations: int = 128,
        num_interpolated_evaluations: int = 128,
        workers: int = -1,
    ):
        """
        Optimizes (maximizes) a given function by alternating between optimisation
        of a proxy function obtrained through interpolation, and exploration of the
        least sampled regions of the optimisation domain.

        Parameters
        ----------
        function: Callable
            Function to optimize.

        bounds: List[Union[Tuple[int, ...], Tuple[float, ...]]]
            Bounds for function parameters

        init_strategies: str
            Initialisation strategies. Can contain: 'corners', 'centers', and 'random'

        exploration_rate: float
            Rate at which to explore

        max_num_evaluations: int
            Maximum number of evaluations of the /a priori/ costly given function.

        num_interpolated_evaluations: int
            Max number of evaluations of the inyterpolated function.

        workers: int
            Number of workers, if -1 the maximum is used.


        Returns
        -------
        optimal_point, optimal_value
        optimal_point: Optimal value for parameters
        optimal_value: Corresponding function value

        """
        # First we figure out the dimensionality of the problem:
        n = len(bounds)

        # Save Function:
        self.function = function

        # Save bounds:
        self.bounds = bounds

        # Second, we allocate the array that stores the evaluations:
        self.x = []
        self.y = []

        # We keep track here of the best evaluation:
        self.best_point = None
        self.best_value = -math.inf

        # First we initialise with some points on the corners:
        if 'corners' in init_strategies:
            with lsection("Evaluating function at corners"):

                if 'centers' in init_strategies:
                    init_grid = tuple((u, 0.5 * (u + v), v) for u, v in bounds)
                else:
                    init_grid = copy(bounds)

                point_list = list(itertools.product(*init_grid))
                self._add_points(point_list, workers=workers, display_points=True)

        # First we initialise with some random points:
        if 'random' in init_strategies:
            with lsection("Evaluating function at random points"):
                point_list = list(self._random_sample() for _ in range(min(4, 2 * n)))
                self._add_points(point_list, workers=workers)

        # Foir how long did we not see an improvement?
        self.since_last_best = 0

        # This is the main loop that evaluates the function:
        with lsection(
            f"Optimizing function with at most {max_num_evaluations} function evaluations within: {bounds}"
        ):
            for i in range(max_num_evaluations):
                # lprint(f"Evaluation #{i}")
                # lprint(f"x={x}")

                # Given the existing points, we can build the interpolating function:
                try:
                    self.interpolator = RBFInterpolator(
                        y=numpy.stack(self.x),
                        d=numpy.stack(self.y),
                        neighbors=8 if len(self.x) < 8 else 4 * n,
                        smoothing=abs(numpy.random.normal(0, 1)) ** 0.5,
                    )

                    # From time to time we just pick points far from all other points:
                    do_explore = numpy.random.random() < exploration_rate

                    # using the interpolator we can quickly search for the best value:
                    new_point = self._delegated_optimizer(
                        do_explore=do_explore,
                        num_evaluations=num_interpolated_evaluations,
                    )

                    # We add that point to the list of points:
                    has_new_best = self._add_points([new_point])

                    lprint(
                        f"{i}{'!' if has_new_best else' '}: {' Exploring' if do_explore else 'Optimizing'}, Best point: {self.best_point}, best value: {self.best_value}, new point: {new_point})"
                    )

                    # Are we running out of patience?
                    if self.since_last_best > patience:
                        # If yes we stop searching:
                        lprint(
                            f"Run out of patience: {self.since_last_best} > {patience} !"
                        )
                        break

                except LinAlgError:
                    lprint("Error while optimizing, let's stop training now!")
                    lprint(f"x={self.x}")
                    traceback.print_exc()
                    break

        lprint(f"Best point: {self.best_point}, best value: {self.best_value}")

        return self.best_point, self.best_value

    def _add_points(self, point_list: List, workers=-1, display_points=False):

        # Normalise points:
        point_list = list(
            numpy.array(point, dtype=numpy.float32) for point in point_list
        )

        def _function(*_point):
            _value = self.function(*_point)
            if display_points:
                lprint(f"New point: {_point} -> {_value}")
            return _value

        # Evaluate function in parallel:
        values = Parallel(n_jobs=workers, backend='threading')(
            delayed(_function)(*point) for point in point_list
        )

        # to return:
        has_new_best = False

        # Going through the list of points:
        for new_value, new_point in zip(values, point_list):

            # Replace NaNs or other weird floats with something better:
            new_value = numpy.nan_to_num(new_value, neginf=-1e6, posinf=-1e6, nan=-1e6)

            # And add this new point to the list:
            self.x.append(new_point)
            self.y.append(new_value)
            # We keep track of the last best evaluation:
            if new_value > self.best_value:
                has_new_best = True
                self.since_last_best = 0
                self.best_value = new_value
                self.best_point = new_point
            else:
                self.since_last_best += 1

        return has_new_best

    def _delegated_optimizer(
        self, do_explore: bool, num_evaluations: int = 128, workers: int = -1
    ):
        # If we ran out of evaluations (recursive call!), then let's return immediately with None:
        if num_evaluations <= 0:
            return None

        # First we figure out the dimensionality of the problem:
        n = len(self.bounds)

        # This is the function to optimize:
        def function(point):

            value = 0

            fallback_exploration = False
            if not do_explore:
                # We compute interpolated value:
                try:
                    # interpolation value:
                    value += self.interpolator(point.reshape(n, -1).T)
                except Exception as e:
                    lprint(f"Exception: {e}")
                    # If there is an issue with interpolation, we fallback on exploration:
                    fallback_exploration = True

            if do_explore or fallback_exploration:
                # point coordinates translated for usage with the kd-tree:
                point_for_tree = numpy.array(point)[numpy.newaxis, ...]

                # We collect neighbors:
                distances, indices = self.interpolator._tree.query(
                    point_for_tree, k=1, workers=workers
                )
                indices = indices.flatten()

                # Corresponding point:
                neighboor = self.x[indices[0]]

                # Vector from neighboor to point:
                vector = neighboor - point

                # We add the lipschitz value to the interpolated value:
                value += norm(vector)

            return value

        if do_explore:
            # Random optimizer is great to avoid getting stuck:
            point = self._random_optimizer(function)
        else:
            # Fast minimisation with Neadler-Mead helps get closer to the optimum,
            # here we don't need randomness as much:
            result = minimize(
                lambda x_: -function(x_),  # this is a minimiser
                x0=self.best_point,
                method='Nelder-Mead',
                bounds=self.bounds,
                options={'maxiter': num_evaluations},
            )
            point = result.x

        # the RBF interpolator hates it when a point occurs multiple times,
        # we clean up:
        while True:
            if _is_in(point, self.x) or point is None:
                # If we get a point we already have, lets pick a point at random:
                # lprint(
                #     f"Point {point} already suggested (or None), trying something else..."
                # )
                try:
                    if point is None:
                        # Best point is None, let's pick a random point:
                        point = self._random_sample()
                    else:
                        # Add noise to current point
                        point = self._add_noise(point, sigma=0.01)
                    # lprint(f"point: {point}")
                except RecursionError:
                    # Fail safe in case we recurse too much:
                    point = self._random_sample()
            else:
                # we are good to go...
                break

        return point

    def _compute_lipschitz(self):
        # point coordinates translated for usage with teh kd-tree:
        point_for_tree = numpy.array(self.best_point)[numpy.newaxis, ...]

        # We collect neighbors:
        distances, indices = self.interpolator._tree.query(
            point_for_tree, k=2 * len(self.best_point), workers=-1
        )
        indices = indices.flatten()
        length = len(indices)

        # enumerate all slops:
        slopes_list = []
        for u in range(0, length - 1):
            for v in range(u + 1, length):
                point_u = self.x[indices[u]]
                point_v = self.x[indices[v]]

                value_u = self.y[indices[u]]
                value_v = self.y[indices[v]]

                slopes = numpy.abs((value_v - value_u) / (point_v - point_u))

                slopes_list.append(slopes)

        # convert to array:
        slopes_array = numpy.stack(slopes_list)

        # We get rid of weird floats:
        slopes_array = numpy.nan_to_num(slopes_array, neginf=0.0, posinf=0.0, nan=0)

        # We compute the median:
        lipschitz = numpy.median(slopes_array, axis=0)

        # make sure that there are no zeroes:
        lipschitz[lipschitz == 0] = numpy.min(lipschitz)

        return lipschitz

    def _random_optimizer(self, function, num_evaluations: int = 128):
        # We keep track of the best values:
        best_point = None
        best_value = -math.inf

        # This is a random optimizer
        for i in range(num_evaluations):
            # We pick a random point:
            point = self._random_sample()

            # evaluate function at point:
            value = function(point)

            # Replace NaNs or other weird floats with something better:
            value = numpy.nan_to_num(value, neginf=-1e6, posinf=-1e6, nan=-1e6)

            # We check if the value is better:
            if value > best_value:
                best_value = value
                best_point = point

        return best_point

    def _random_sample(self):
        # First we figure out the dimensionality of the problem:
        n = len(self.bounds)

        # List to store coordinates
        point = []

        # Loop through each coordinate:
        for i in range(n):
            min_r, max_r = self.bounds[i]

            # int coodinates:
            if type(min_r) is int or type(min_r) in [numpy.int32, numpy.int64]:
                coord = numpy.random.randint(min_r, max_r)

            # float coordinates:
            elif type(min_r) is float or type(min_r) in [numpy.float32, numpy.float64]:
                coord = numpy.random.uniform(min_r, max_r)

            point.append(coord)

        return numpy.array(point)

    def _add_noise(self, point, sigma: float = 1e-6):
        # range widths:
        widths = numpy.array(tuple(abs(float(u - v)) for u, v in self.bounds))

        # mins and maxes :
        mins = numpy.array(tuple(float(min(u, v)) for u, v in self.bounds))
        maxs = numpy.array(tuple(float(max(u, v)) for u, v in self.bounds))

        # List to store coordinates
        point += numpy.random.normal(0, sigma, point.shape) * 0.5 * widths

        # clip to range:
        point = numpy.maximum(point, mins)
        point = numpy.minimum(point, maxs)

        return numpy.array(point)


def _is_in(array, array_list):
    for a in array_list:
        # RBF interpolator hates it when two coordinates of two points have the same values...
        if (a == array).any():
            return True

    return False



# flake8: noqa
import time
from datetime import date
from os.path import join

import numpy

from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io import io
from aydin.io.datasets import examples_single
from aydin.io.folders import get_temp_folder
from aydin.io.io import mapped_tiff
from aydin.it.fgr import ImageTranslatorFGR
from aydin.regression.nn import NNRegressor


def demo():
    today = date.today()
    filename = f"result_{today.strftime('%m_%d_%Y_%H_%M_%S')}.tiff"

    # (3, 320, 865, 1014)
    image_path = examples_single.cognet_nanotube_200fps.get_path()
    array, metadata = io.imread(image_path)
    print(array.shape)
    train = array  # [0:120]
    infer = array  # [0:120]

    # print(f"Number of distinct features in image: {len(numpy.unique(infer))}")

    print(f"train: {train.shape}, inference:{infer.shape} ")

    batch_dims = (False, False, False)

    generator = StandardFeatureGenerator(  # kernel_widths=[3, 3, 1, 1, 1,  1,  1,  1,   1,   1],
        # kernel_scales=[0, 1, 2, 3, 7, 15, 31, 63, 127, 255],
        # kernel_shapes=['li'] * 2 + ['l1'] * 8,
        max_level=10,
        dtype=numpy.float16,
        include_scale_one=False,
        include_spatial_features=True,
    )

    regressor = NNRegressor(max_epochs=200, patience=15)

    it = ImageTranslatorFGR(generator, regressor)

    start = time.time()
    it.train(train, train, batch_axes=batch_dims)
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")

    output_file = join(get_temp_folder(), filename)

    print(f"Output file: {output_file}")

    # We write the stack to a temp file:
    with mapped_tiff(output_file, infer.shape, infer.dtype) as denoised_tiff:
        # denoised = offcore_array(whole.shape, whole.dtype)

        start = time.time()
        denoised = it.translate(
            infer, translated_image=denoised_tiff, batch_axes=batch_dims, tile_size=512
        )
        stop = time.time()

        print(f"Writing to file: {output_file} ")
        denoised_tiff[...] = denoised[...]

        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(train, name='train')
            viewer.add_image(infer, name='infer')
            print(f"Inference: elapsed time:  {stop - start} ")
            viewer.add_image(denoised, name='denoised')


demo()



# flake8: noqa
import numpy
import numpy as np
from skimage.data import camera

from aydin.io.datasets import (
    normalise,
    add_noise,
    dots,
    lizard,
    pollen,
    newyork,
    characters,
    examples_single,
)
from aydin.util.crop.rep_crop import representative_crop
from aydin.util.log.log import Log, lsection


def demo_representative_crop(image, fast_mode: bool = False, display: bool = True):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    Log.enable_output = True
    Log.set_log_max_depth(5)

    with lsection(f"Computing crop for image of shape: {image.shape}"):
        crop_size = 64000
        crop = representative_crop(
            image, crop_size=crop_size, fast_mode=fast_mode, display_crop=False
        )

    if display:
        import napari

        with napari.gui_qt():
            viewer = napari.Viewer()
            viewer.add_image(image, name='image')
            viewer.add_image(crop, name='crop')

    assert crop.size <= int(crop_size * 1.05)


if __name__ == "__main__":

    islet = examples_single.royerlab_hcr.get_array().squeeze()[1]
    demo_representative_crop(islet, fast_mode=True)

    demo_representative_crop(newyork())
    demo_representative_crop(camera())
    demo_representative_crop(characters())
    demo_representative_crop(pollen())
    demo_representative_crop(lizard())
    demo_representative_crop(dots())



import numpy

from aydin.analysis.camera_simulation import simulate_camera_image
from aydin.analysis.resolution_estimate import resolution_estimate
from aydin.analysis.snr_estimate import snr_estimate
from aydin.io.datasets import camera
from aydin.util.log.log import lprint, Log


def demo_camera_simulation():
    Log.enable_output = True

    clean_image = camera()
    noisy_video = numpy.stack(
        (
            simulate_camera_image(clean_image, exposure_time_s=e / 32)
            for e in range(1, 32)
        )
    )

    fc, _ = resolution_estimate(noisy_video[16])
    snr = snr_estimate(noisy_video[16])

    lprint(f"Resolution: {fc}, snr: {snr}")

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(clean_image, name='clean_image')
        viewer.add_image(noisy_video, name='noisy_video')


def demo_camera_simulation_video():
    clean_image = camera()
    noisy_video = numpy.stack(
        (simulate_camera_image(clean_image // 32) for _ in range(64))
    )

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(clean_image, name='clean_image')
        viewer.add_image(noisy_video, name='noisy_video')


demo_camera_simulation()
demo_camera_simulation_video()



import math
from collections import OrderedDict
from itertools import chain

import numpy
import torch
from aydin.it.base import ImageTranslatorBase
from aydin.nn.pytorch.models.jidcnet import JInet2D
from aydin.nn.pytorch.models.masking import Masking
from aydin.nn.pytorch.optimizers.esadam import ESAdam
from aydin.util.array.nd import extract_tiles
from aydin.util.log.log import lsection, lprint
from torch import nn
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.utils.data import Dataset


def to_numpy(tensor):
    return tensor.clone().detach().cpu().numpy()


class PTCNNImageTranslator(ImageTranslatorBase):
    """
    Pytorch-based CNN image translator
    """

    def __init__(
        self,
        max_epochs=1024,
        patience=128,
        patience_epsilon=0.0,
        learning_rate=0.01,
        batch_size=64,
        model_class=JInet2D,
        denoise_loss='l1',
        deconv_loss='l1',
        normaliser_type='percentile',
        balance_training_data=None,
        keep_ratio=1,
        max_voxels_for_training=4e6,
        use_cuda=True,
        device_index=0,
    ):
        """
        Constructs an image translator using the pytorch deep learning library.

        :param normaliser_type: normaliser type
        :param balance_training_data: balance data ? (limits number training entries per target value histogram bin)

        """
        super().__init__(normaliser_type)

        use_cuda = use_cuda and (torch.cuda.device_count() > 0)
        self.device = torch.device(f"cuda:{device_index}" if use_cuda else "cpu")
        lprint(f"Using device: {self.device}")

        self.max_epochs = max_epochs
        self.patience = patience
        self.patience_epsilon = patience_epsilon
        self.learning_rate = learning_rate
        self.batch_size = batch_size
        self.denoise_loss = denoise_loss
        self.deconv_loss = deconv_loss
        self.max_voxels_for_training = max_voxels_for_training
        self.keep_ratio = keep_ratio
        self.balance_training_data = balance_training_data

        self.model_class = model_class

        # These parametres are 'private'  we don't expect the user to modify them,
        # but since they are fields they can be modified if need be after construction...
        self.l1_weight_regularisation = 1e-9
        self.l2_weight_regularisation = 1e-9
        self.training_noise = 0.001
        self.reload_best_model_period = max_epochs  # //2
        self.reduce_lr_patience = patience // 2  # TODO: consider patience//2
        self.reduce_lr_factor = 0.5
        self.masking = False
        self.enforce_blind_spot = True
        self.optimizer_class = ESAdam
        self.max_tile_size = 1024  # TODO: adjust based on available memory

        self._stop_training_flag = False

        # pytorch_info = torch.__config__.show().replace('\n', ';')
        # lprint(f"PyTorch Info: {pytorch_info}")

    def save(self, path: str):
        """
        Saves a 'all-batteries-included' image translation model at a given path (folder).
        :param path: path to save to
        """
        with lsection(f"Saving 'classic' image translator to {path}"):
            frozen = super().save(path)
            pass
            # TODO: complete!

        return frozen

    def _load_internals(self, path: str):
        with lsection(f"Loading 'classic' image translator from {path}"):
            pass
            # TODO: complete!

    # We exclude certain fields from saving:
    def __getstate__(self):
        state = self.__dict__.copy()
        del state['XXXXXXX']
        return state

    def stop_training(self):
        lprint("Received notification to stop training loop now.")
        self._stop_training_flag = True

    def _train(
        self,
        input_image,
        target_image,
        train_valid_ratio=0.1,
        callback_period=3,
        jinv=False,
    ):
        self._stop_training_flag = False

        shape = input_image.shape
        # num_batches = shape[0]
        num_input_channels = input_image.shape[1]
        num_output_channels = target_image.shape[1]
        # num_spatiotemp_dim = input_image.ndim - 2

        # tile size:
        tile_size = min(self.max_tile_size, min(shape[2:]))

        # Decide on how many voxels to be used for validation:
        num_val_voxels = int(train_valid_ratio * input_image.size)
        lprint(
            f"Number of voxels used for validation: {num_val_voxels} (train_valid_ratio={train_valid_ratio})"
        )

        # Generate random coordinates for these voxels:
        val_voxels = tuple(numpy.random.randint(d, size=num_val_voxels) for d in shape)
        lprint(f"Validation voxel coordinates: {val_voxels}")

        # Training Tile size:
        lprint(f"Train Tile dimensions: {tile_size}")

        # Prepare Training Dataset:
        dataset = self._get_dataset(
            input_image,
            target_image,
            self.self_supervised,
            tilesize=tile_size,
            mode='grid',
            validation_voxels=val_voxels,
        )
        lprint(f"Number tiles for training: {len(dataset)}")

        # Training Data Loader:
        # num_workers = max(3, os.cpu_count() // 2)
        num_workers = 0  # faster if data is already in memory...
        lprint(f"Number of workers for loading training/validation data: {num_workers}")
        data_loader = torch.utils.data.DataLoader(
            dataset,
            batch_size=1,  # self.batch_size,
            shuffle=True,
            num_workers=num_workers,
            pin_memory=True,
        )

        # Model
        self.model = self.model_class(num_input_channels, num_output_channels).to(
            self.device
        )

        number_of_parameters = sum(
            p.numel() for p in self.model.parameters() if p.requires_grad
        )
        lprint(
            f"Number of trainable parameters in {self.model_class} model: {number_of_parameters}"
        )

        if self.masking:
            self.masked_model = Masking(self.model).to(self.device)

        lprint(f"Optimizer class: {self.optimizer_class}")
        lprint(f"Learning rate : {self.learning_rate}")

        # Optimizer:
        optimizer = self.optimizer_class(
            chain(self.model.parameters()),
            lr=self.learning_rate,
            start_noise_level=self.training_noise,
            weight_decay=self.l2_weight_regularisation,
        )

        lprint(f"Optimizer: {optimizer}")

        # Denoise loss functon:
        loss_function = nn.L1Loss()
        if self.denoise_loss.lower() == 'l2':
            lprint("Training/Validation loss: L2")
            if self.masking:
                loss_function = (
                    lambda u, v, m: (u - v) ** 2 if m is None else ((u - v) * m) ** 2
                )
            else:
                loss_function = lambda u, v: (u - v) ** 2  # noqa : E731

        elif self.denoise_loss.lower() == 'l1':
            if self.masking:
                loss_function = (
                    lambda u, v, m: torch.abs(u - v)
                    if m is None
                    else torch.abs((u - v) * m)
                )
            else:
                loss_function = lambda u, v: torch.abs(u - v)  # noqa : E731
            lprint("Training/Validation loss: L1")

        # Start training:
        self._train_loop(data_loader, optimizer, loss_function)

    def _get_dataset(
        self,
        input_image: numpy.ndarray,
        target_image: numpy.ndarray,
        self_supervised: bool,
        tilesize: int,
        mode: str,
        validation_voxels,
    ):
        class _Dataset(Dataset):
            def __init__(self, input_image, target_image, tilesize):
                """ """

                num_channels_input = input_image.shape[1]
                num_channels_target = target_image.shape[1]

                def extract(image):
                    return extract_tiles(
                        image,
                        tile_size=tilesize,
                        extraction_step=tilesize,
                        flatten=True,
                    )

                bc_flat_input_image = input_image.reshape(-1, *input_image.shape[2:])
                bc_flat_input_tiles = numpy.concatenate(
                    [extract(x) for x in bc_flat_input_image]
                )
                self.input_tiles = bc_flat_input_tiles.reshape(
                    -1, num_channels_input, *bc_flat_input_tiles.shape[1:]
                )

                if self_supervised:
                    self.target_tiles = self.input_tiles
                else:
                    bc_flat_target_image = target_image.reshape(
                        -1, *target_image.shape[2:]
                    )
                    bc_flat_target_tiles = numpy.concatenate(
                        [extract(x) for x in bc_flat_target_image]
                    )
                    self.target_tiles = bc_flat_target_tiles.reshape(
                        -1, num_channels_target, *bc_flat_target_tiles.shape[1:]
                    )

                mask_image = numpy.zeros_like(input_image)
                mask_image[validation_voxels] = 1

                bc_flat_mask_image = mask_image.reshape(-1, *mask_image.shape[2:])
                bc_flat_mask_tiles = numpy.concatenate(
                    [extract(x) for x in bc_flat_mask_image]
                )
                self.mask_tiles = bc_flat_mask_tiles.reshape(
                    -1, num_channels_input, *bc_flat_mask_tiles.shape[1:]
                )

            def __len__(self):
                return len(self.input_tiles)

            def __getitem__(self, index):
                input = self.input_tiles[index, ...]
                target = self.target_tiles[index, ...]
                mask = self.mask_tiles[index, ...]

                return (input, target, mask)

        if mode == 'grid':
            return _Dataset(input_image, target_image, tilesize)
        else:
            return None

    def _train_loop(self, data_loader, optimizer, loss_function):

        # Scheduler:
        scheduler = ReduceLROnPlateau(
            optimizer,
            'min',
            factor=self.reduce_lr_factor,
            verbose=True,
            patience=self.reduce_lr_patience,
        )

        best_val_loss_value = math.inf
        best_model_state_dict = None
        patience_counter = 0

        with lsection("Training loop:"):
            lprint(f"Maximum number of epochs: {self.max_epochs}")
            lprint(
                f"Training type: {'self-supervised' if self.self_supervised else 'supervised'}"
            )

            for epoch in range(self.max_epochs):
                with lsection(f"Epoch {epoch}:"):

                    train_loss_value = 0
                    val_loss_value = 0
                    iteration = 0
                    for i, (input_images, target_images, val_mask_images) in enumerate(
                        data_loader
                    ):
                        lprint(f"index: {i}, shape:{input_images.shape}")

                        input_images_gpu = input_images.to(
                            self.device, non_blocking=True
                        )
                        target_images_gpu = target_images.to(
                            self.device, non_blocking=True
                        )
                        validation_mask_images_gpu = val_mask_images.to(
                            self.device, non_blocking=True
                        )

                        # Adding training noise to input:
                        if self.training_noise > 0:
                            with torch.no_grad():
                                alpha = self.training_noise / (
                                    1 + (10000 * epoch / self.max_epochs)
                                )
                                lprint(f"Training noise level: {alpha}")
                                training_noise = alpha * torch.randn_like(input_images)
                                input_images_gpu += training_noise.to(
                                    input_images_gpu.device
                                )

                        # Clear gradients w.r.t. parameters
                        optimizer.zero_grad()

                        # Forward pass:
                        self.model.train()
                        if self.masking:
                            translated_images_gpu = self.masked_model(input_images_gpu)
                        else:
                            translated_images_gpu = self.model(input_images_gpu)

                        # apply forward model:
                        forward_model_images_gpu = self._forward_model(
                            translated_images_gpu
                        )

                        # validation masking:
                        u = forward_model_images_gpu * (1 - validation_mask_images_gpu)
                        v = target_images_gpu * (1 - validation_mask_images_gpu)

                        # with napari.gui_qt():
                        #     viewer = napari.Viewer()
                        #     viewer.add_image(to_numpy(validation_mask_images_gpu), name='validation_mask_images_gpu')
                        #     viewer.add_image(to_numpy(forward_model_images_gpu), name='forward_model_images_gpu')
                        #     viewer.add_image(to_numpy(target_images_gpu), name='target_images_gpu')

                        # translation loss (per voxel):
                        if self.masking:
                            mask = self.masked_model.get_mask()
                            translation_loss = loss_function(u, v, mask)
                        else:
                            translation_loss = loss_function(u, v)

                        # loss value (for all voxels):
                        translation_loss_value = translation_loss.mean()

                        # Additional losses:
                        additional_loss_value = self._additional_losses(
                            translated_images_gpu, forward_model_images_gpu
                        )
                        if additional_loss_value is not None:
                            translation_loss_value += additional_loss_value

                        # backpropagation:
                        translation_loss_value.backward()

                        # Updating parameters
                        optimizer.step()

                        # post optimisation -- if needed:
                        self.model.post_optimisation()

                        # If self-supervised then enforce blind-spot:
                        if self.self_supervised and self.enforce_blind_spot:
                            try:
                                self.model.enforce_blind_spot()
                                lprint("Post optimisation corrections applied to model")
                            except AttributeError:
                                lprint(
                                    "NO post optimisation corrections applied to model"
                                )

                        # update training loss_deconvolution for whole image:
                        train_loss_value += translation_loss_value.item()
                        iteration += 1

                        # Validation:
                        with torch.no_grad():
                            # Forward pass:
                            self.model.eval()
                            if self.masking:
                                translated_images_gpu = self.masked_model(
                                    input_images_gpu
                                )
                            else:
                                translated_images_gpu = self.model(input_images_gpu)

                            # apply forward model:
                            forward_model_images_gpu = self._forward_model(
                                translated_images_gpu
                            )

                            # validation masking:
                            u = forward_model_images_gpu * validation_mask_images_gpu
                            v = target_images_gpu * validation_mask_images_gpu

                            # translation loss (per voxel):
                            if self.masking:
                                translation_loss = loss_function(u, v, None)
                            else:
                                translation_loss = loss_function(u, v)

                            # loss values:
                            translation_loss_value = (
                                translation_loss.mean().cpu().item()
                            )

                            # update validation loss_deconvolution for whole image:
                            val_loss_value += translation_loss_value
                            iteration += 1

                    train_loss_value /= iteration
                    lprint("Training loss value: {train_loss_value}")

                    val_loss_value /= iteration
                    lprint("Validation loss value: {val_loss_value}")

                    # Learning rate schedule:
                    scheduler.step(val_loss_value)

                    if val_loss_value < best_val_loss_value:
                        lprint("## New best val loss!")
                        if val_loss_value < best_val_loss_value - self.patience_epsilon:
                            lprint("## Good enough to reset patience!")
                            patience_counter = 0

                        # Update best val loss value:
                        best_val_loss_value = val_loss_value

                        # Save model:
                        best_model_state_dict = OrderedDict(
                            {k: v.to('cpu') for k, v in self.model.state_dict().items()}
                        )

                    else:
                        if (
                            epoch % max(1, self.reload_best_model_period) == 0
                            and best_model_state_dict
                        ):
                            lprint("Reloading best models to date!")
                            self.model.load_state_dict(best_model_state_dict)

                        if patience_counter > self.patience:
                            lprint("Early stopping!")
                            break

                        # No improvement:
                        lprint(
                            "No improvement of validation losses, patience = {patience_counter}/{self.patience} "
                        )
                        patience_counter += 1

                    lprint("## Best val loss: {best_val_loss_value}")

                    if self._stop_training_flag:
                        lprint("Training interupted!")
                        break

        lprint("Reloading best models to date!")
        self.model.load_state_dict(best_model_state_dict)

        if self.self_supervised and self.enforce_blind_spot:
            try:
                self.model.fill_blind_spot()
                lprint("Blind spot filled!")
            except AttributeError:
                lprint("Blind spot NOT filled! (no method available)")

    def _additional_losses(self, translated_image, forward_model_image):
        return None

    def _forward_model(self, input):
        return input

    def _translate(self, input_image, image_slice=None, whole_image_shape=None):
        """Internal method that translates an input image on the basis of the trained model.

        Parameters
        ----------
        input_image
            input image
        image_slice
        whole_image_shape

        Returns
        -------

        """
        input_image = torch.Tensor(input_image)
        input_image = input_image.to(self.device)
        inferred_image: torch.Tensor = self.model(input_image)
        inferred_image = inferred_image.detach().cpu().numpy()
        return inferred_image

    def visualise_weights(self):
        try:
            self.model.visualise_weights()
        except AttributeError:
            lprint(
                "Method 'visualise_weights()' unavailable, cannot visualise weights. "
            )



from abc import ABC, abstractmethod

# from numpy.typing import ArrayLike


class ImageTransformBase(ABC):
    """Image Transforms base class

    Transforms take an array, apply a transformation that is meant to
    facilitate denoising (or another image restoration task), and then
    optionally, offer a way to go back by undoing the transformation. For
    example, motion correction can be applied to facilitate denoising,
    and then after denoising has ben performed, the motion can be reapplied,
    if really needed.
    """

    preprocess_description = " (before denoising)"
    postprocess_description = " (after denoising)"
    postprocess_supported = None

    def __init__(self, priority: float = -1, do_postprocess: bool = True, **kwargs):
        """

        Parameters
        ----------
        priority : float
            The priority is a value within [0,1] used to determine the order in
            which to apply the pre- and post-processing transforms. Transforms
            are sorted and applied in ascending order during preprocesing and in
            the reverse, descending, order during post-processing.

        do_postprocess : bool When True, post-processing will occur,
        when False, no post-processing is done.

        kwargs
        """
        self.priority = priority
        self.do_postprocess = do_postprocess

    @abstractmethod
    def preprocess(self, array):
        """ """
        raise NotImplementedError()

    @abstractmethod
    def postprocess(self, array):
        """ """
        raise NotImplementedError()



import functools
import numpy


class memoize_last(object):
    """Decorator class used to cache the most recent value of a function
    or method based on the signature of its arguments. If any single
    argument changes, the function or method is evaluated afresh.

    Importantly: works even for numpy arrays although they are not hashable.
    We use the object id, so great care has to be taken when using this!!

    Original implementation from:
    https://gist.github.com/dpo/1222577
    https://github.com/numpy/numpy/issues/14294

    """

    def __init__(self, callable):
        self._callable = callable
        self._callable_is_method = False
        self.value = None  # Cached value or derivative.
        self._args_hashes = {}
        return

    def __get_hash(self, x):
        # Return signature of argument.
        if isinstance(x, numpy.ndarray):
            # _x = x.view(numpy.uint8)
            # return hash(hashlib.sha1(_x).hexdigest())
            return id(x)

        return hash(x)

    def __call__(self, *args, **kwargs):
        # The callable will be called if any single argument is new or changed.

        callable = self._callable
        evaluate = False

        # If we're memoizing a class method, the first argument will be 'self'
        # and need not be memoized.
        firstarg = 1 if self._callable_is_method else 0

        # Get signature of all arguments.
        nargs = callable.__code__.co_argcount  # Non-keyword arguments.
        argnames = callable.__code__.co_varnames[firstarg:nargs]
        argvals = args[firstarg:]

        allargs = dict(zip(argnames, argvals)) | kwargs

        for (argname, argval) in allargs.items():

            _arg_hash = self.__get_hash(argval)

            try:
                cached_arg_hash = self._args_hashes[argname]
                if cached_arg_hash != _arg_hash:
                    self._args_hashes[argname] = _arg_hash
                    evaluate = True

            except KeyError:
                self._args_hashes[argname] = _arg_hash
                evaluate = True

        # If cache invalid, recompute value:
        if evaluate:
            self.value = callable(*args, **kwargs)

        print(f"{tuple(f'{k}:{id(v)}' for k,v in allargs.items())} -> {id(self.value)}")

        return self.value

    def __get__(self, obj, objtype):
        "Support instance methods."
        self._callable_is_method = True
        return functools.partial(self.__call__, obj)

    def __repr__(self):
        "Return the wrapped function or method's docstring."
        return self.method.__doc__



from aydin.util.misc.combinatorics import closest_product


def test_closest_product():

    u = [1, 2, 5, 7, 9, 10]

    N = 15
    result = closest_product(u, N)
    print(f"closest_product({u}, {N}) = {result}")
    assert result == [1, 3]

    N = 27
    result = closest_product(u, N)
    print(f"closest_product({u}, {N}) = {result}")
    assert result is None



import numpy
from scipy.ndimage import uniform_filter


def scipy_uniform_filter(
    image, size=3, output=None, mode="nearest", cval=0.0, origin=0
):
    # Save original image dtype:
    original_dtype = image.dtype

    # Scipy does not support float16 yet:
    dtype = numpy.float32 if original_dtype == numpy.float16 else original_dtype
    image = image.astype(dtype=dtype, copy=False)

    output = uniform_filter(image, size=size, mode=mode, cval=cval, origin=origin)

    output = output.astype(dtype=original_dtype, copy=False)

    return output



# flake8: noqa
import time

import numpy as np

from aydin.analysis.camera_simulation import simulate_camera_image
from aydin.features.standard_features import StandardFeatureGenerator
from aydin.io.datasets import newyork, characters, pollen
from aydin.it.fgr import ImageTranslatorFGR
from aydin.it.transforms.highpass import HighpassTransform
from aydin.it.transforms.range import RangeTransform
from aydin.it.transforms.variance_stabilisation import VarianceStabilisationTransform
from aydin.regression.cb import CBRegressor
from aydin.util.log.log import Log


def demo(image, name):
    """
    Demo for self-supervised denoising using camera image with synthetic noise
    """
    Log.enable_output = True
    Log.set_log_max_depth(7)

    image = image.astype(np.float32) * 0.1
    noisy = simulate_camera_image(image)

    import napari

    with napari.gui_qt():
        viewer = napari.Viewer()
        viewer.add_image(image, name='image')
        viewer.add_image(noisy, name='noisy')

        for mode in [
            'identity',
            'box-cox',
            'yeo-johnson',
            'quantile',
            'anscomb',
            'log',
            'sqrt',
        ]:
            denoised = _denoise(noisy, mode)
            viewer.add_image(denoised, name=f'denoised_{mode}')


def _denoise(noisy, mode):
    generator = StandardFeatureGenerator(
        include_corner_features=True,
        include_scale_one=False,
        include_fine_features=True,
        include_spatial_features=True,
    )
    regressor = CBRegressor(patience=32, gpu=True, min_num_estimators=1024)
    it = ImageTranslatorFGR(feature_generator=generator, regressor=regressor)
    it.transforms_list.append(VarianceStabilisationTransform(mode=mode))
    it.transforms_list.append(RangeTransform())
    it.transforms_list.append(HighpassTransform())
    # it.transforms_list.append(PaddingTransform())
    print("training starts")
    start = time.time()
    it.train(noisy, noisy)
    stop = time.time()
    print(f"Training: elapsed time:  {stop - start} ")
    # in case of batching we have to do this:
    start = time.time()
    denoised = it.translate(noisy)
    stop = time.time()
    print(f"inference: elapsed time:  {stop - start} ")
    return denoised


# lizard_image = lizard()
# demo(lizard_image, "lizard")
pollen_image = pollen()
demo(pollen_image, "pollen")
newyork_image = newyork()
demo(newyork_image, "newyork")
characters_image = characters()
demo(characters_image, "characters")


# dots_image = dots()
# demo(dots_image, "dots")
